from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.evaluation.bleu_tool.bleu import Bleu


class BLEU(EvaluationMetricBase):
    """
        The BLEU evaluation metric class.
    Parameters
    ----------
    n_grams: list[int]
        The BLEU's n_gram parameter. The results will be returned according to the ``n_grams`` one-by-one.
    verbase: int, default = 0
        The log indicator. If set to 0, it will output no logs.
    """

    def __init__(self, n_grams, verbase=0):
        super(BLEU, self).__init__()
        max_gram = self._check_available(n_grams)
        self.scorer = Bleu(max_gram, verbase=verbase)
        self.metrics = n_grams

    def calculate_scores(self, ground_truth, predict):
        """
            The BLEU calculation function. It will compute the BLEU scores.
        Parameters
        ----------
        ground_truth: list[string]
            The ground truth (correct) target values. It is a list of strings.
        predict: list[string]
            The predicted target values. It is a list of strings.
        Returns
        -------
        score: list[float]
            The list contains BLEU_n results according to ``n_grams``.
        scores: list[list[float]]
            The specific results for each needed BLEU_n metric.
        """
        ref_list = [list(map(str.strip, refs)) for refs in zip(ground_truth)]

        refs = {idx: strippedlines for (idx, strippedlines) in enumerate(ref_list)}
        hyps = {idx: [lines.strip()] for (idx, lines) in enumerate(predict)}
        assert len(refs) == len(hyps)
        score, scores = self.scorer.compute_score(refs, hyps)
        score_ret = [score[i - 1] for i in self.metrics]
        scores_ret = [scores[i - 1] for i in self.metrics]
        return score_ret, scores_ret

    @staticmethod
    def _check_available(n_grams):
        """
            The function to check the parameters.
            If all tests are passed, it will find the max value in ``n_grams``.
        Parameters
        ----------
        n_grams: list[int]

        Returns
        -------
        max_n_grams_value: int
        """
        n_grams_ok = True
        if isinstance(n_grams, list):
            for i in n_grams:
                if not isinstance(i, int):
                    n_grams_ok = False
        else:
            n_grams_ok = False
        if not n_grams_ok:
            raise TypeError("argument n_grams must be in list of int.")
        return max(n_grams)



import torch
import torch.nn as nn
import torch.nn.functional as F
import dgl.function as fn
from dgl.nn.pytorch import SAGEConv
from dgl.utils import expand_as_pair
from .base import GNNLayerBase, GNNBase
from dgl.utils import expand_as_pair,check_eq_shape
import warnings

class GraphSAGE(GNNBase):
    r"""Multi-layered `GraphSAGE Network <https://arxiv.org/pdf/1706.02216.pdf>`__
    Support both unidirectional (i.e., regular) and bidirectional (i.e., `bi_sep` and `bi_fuse`) versions.
    
    .. math::
        h_{\mathcal{N}(i)}^{(l+1)} & = \mathrm{aggregate}
        \left(\{h_{j}^{l}, \forall j \in \mathcal{N}(i) \}\right)
        h_{i}^{(l+1)} & = \sigma \left(W \cdot \mathrm{concat}
        (h_{i}^{l}, h_{\mathcal{N}(i)}^{l+1} + b) \right)
        h_{i}^{(l+1)} & = \mathrm{norm}(h_{i}^{l})
        
    Parameters
    ----------
    num_layers: int
        Number of GraphSAGE layers.
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
        If aggregator type is ``gcn``, the feature size of source and destination nodes
        are required to be the same.
    hidden_size: int list of int
        Hidden layer size.
        If a scalar is given, the sizes of all the hidden layers are the same.
        If a list of scalar is given, each element in the list is the size of each hidden layer.
        Example: [100,50]
    output_size : int
        Output feature size. For the bisep version, the output size is 2*output_size
    direction_option: str
        Whether use unidirectional (i.e., regular) or bidirectional (i.e., `bi_sep` and `bi_fuse`) versions.
    feat_drop : float, optional
        Dropout rate on feature, defaults: ``0``.
    aggregator_type : str
        Aggregator type to use (``mean``, ``gcn``, ``pool``, ``lstm``).
    bias : bool
        If True, adds a learnable bias to the output. Default: ``True``.
    norm : callable activation function/layer or None, optional
        If not None, applies normalization to the updated node features.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                num_layers,
                input_size,
                hidden_size,
                output_size,
                aggregator_type,
                direction_option='undirected',
                feat_drop=0.,
                bias=True,
                norm=None,
                activation=None,
                use_edge_weight=False):
        super(GraphSAGE, self).__init__()
        self.use_edge_weight=use_edge_weight
        self.num_layers = num_layers
        self.direction_option = direction_option
        self.GraphSAGE_layers = nn.ModuleList()

        #transform the hidden size format
        if self.num_layers>1 and type(hidden_size) is int:
            hidden_size=[hidden_size for i in range(self.num_layers-1)]


        if self.num_layers>1:
            # input projection
            self.GraphSAGE_layers.append(GraphSAGELayer(input_size,
                                            hidden_size[0],
                                            aggregator_type,
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            bias=bias,
                                            norm=norm,
                                            activation=activation))
        # hidden layers
        for l in range(1, self.num_layers - 1):
            # due to multi-head, the input_size = hidden_size * num_heads
            self.GraphSAGE_layers.append(GraphSAGELayer(hidden_size[l - 1],
                                            hidden_size[l],
                                            aggregator_type,
                                            direction_option=self.direction_option,
                                            feat_drop=feat_drop,
                                            bias=bias,
                                            norm=norm,
                                            activation=activation))
        # output projection
        self.GraphSAGE_layers.append(GraphSAGELayer(hidden_size[-1] if self.num_layers > 1 else input_size,
                                        output_size,
                                        aggregator_type,
                                        direction_option=self.direction_option,
                                        feat_drop=feat_drop,
                                        bias=bias,
                                        norm=norm,
                                        activation=activation))

    def forward(self, graph):
        r"""Compute GraphSAGE layer.
        
        Parameters
        ----------
        graph : GraphData
            The graph with node feature stored in the feature field named as
            "node_feat".
            The node features are used for message passing.
        
        Returns
        -------
        graph : GraphData
            The graph with generated node embedding stored in the feature field
            named as "node_emb".
        """

        h=graph.node_features['node_feat'] #get the node feature tensor from graph
        g = graph.to_dgl() #transfer the current NLPgraph to DGL graph
        edge_weight=None
        reverse_edge_weight=None
        if self.use_edge_weight==True:
            edge_weight = graph.edge_features['edge_weight']
            reverse_edge_weight = graph.edge_features['reverse_edge_weight']
        # output projection
        if self.num_layers>1:
          for l in range(0,self.num_layers - 1):
              h = self.GraphSAGE_layers[l](g, h, edge_weight,reverse_edge_weight)

        logits = self.GraphSAGE_layers[-1](g, h, edge_weight,reverse_edge_weight)

        if self.direction_option == 'bi_sep':
            logits = torch.cat(logits, -1)
        else:
            logits = logits

        graph.node_features['node_emb']=logits #put the results into the NLPGraph

        return graph



class GraphSAGELayer(GNNLayerBase):
    r"""A unified wrapper for `GraphSAGE Network <https://arxiv.org/pdf/1706.02216.pdf>`__
    Support both unidirectional (i.e., regular) and bidirectional (i.e., `bi_sep` and `bi_fuse`) versions.
    
    .. math::
        h_{\mathcal{N}(i)}^{(l+1)} & = \mathrm{aggregate}
        \left(\{h_{j}^{l}, \forall j \in \mathcal{N}(i) \}\right)
        h_{i}^{(l+1)} & = \sigma \left(W \cdot \mathrm{concat}
        (h_{i}^{l}, h_{\mathcal{N}(i)}^{l+1} + b) \right)
        h_{i}^{(l+1)} & = \mathrm{norm}(h_{i}^{l})
    
    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied to a unidirectional bipartite graph, ``input_size``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
    output_size : int
        Output feature size.
    direction_option: str
        Whether use unidirectional (i.e., 'undirected') or bidirectional (i.e., `bi_sep` and `bi_fuse`) versions.
    feat_drop : float, optional
        Dropout rate on feature, defaults: ``0``.
    aggregator_type : str
        Aggregator type to use (``mean``, ``gcn``, ``pool``, ``lstm``).
    bias : bool
        If True, adds a learnable bias to the output. Default: ``True``.
    norm : callable activation function/layer or None, optional
        If not None, applies normalization to the updated node features.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                input_size,
                output_size,
                aggregator_type,
                direction_option='undirected',
                feat_drop=0.,
                bias=True,
                norm=None,
                activation=None):
        super(GraphSAGELayer, self).__init__()
        if direction_option == 'undirected':
            self.model = UndirectedGraphSAGELayerConv(input_size,
                                        output_size,
                                        aggregator_type,
                                        feat_drop=feat_drop,
                                        bias=bias,
                                        norm=norm,
                                        activation=activation)
        elif direction_option == 'bi_sep':
            self.model = BiSepGraphSAGELayerConv(input_size,
                                        output_size,
                                        aggregator_type,
                                        feat_drop=feat_drop,
                                        bias=bias,
                                        norm=norm,
                                        activation=activation)
        elif direction_option == 'bi_fuse':
            self.model = BiFuseGraphSAGELayerConv(input_size,
                                        output_size,
                                        aggregator_type,
                                        feat_drop=feat_drop,
                                        bias=bias,
                                        norm=norm,
                                        activation=activation)
        else:
            raise RuntimeError('Unknown `direction_option` value: {}'.format(direction_option))

    def forward(self, graph, feat, edge_weight=None,reverse_edge_weight=None):
        return self.model(graph, feat, edge_weight,reverse_edge_weight)

class UndirectedGraphSAGELayerConv(GNNLayerBase):
    r"""GraphSAGE layer from paper `Inductive Representation Learning on
    Large Graphs <https://arxiv.org/pdf/1706.02216.pdf>`__.
    
    .. math::
        h_{\mathcal{N}(i)}^{(l+1)} & = \mathrm{aggregate}
        \left(\{h_{j}^{l}, \forall j \in \mathcal{N}(i) \}\right)
        h_{i}^{(l+1)} & = \sigma \left(W \cdot \mathrm{concat}
        (h_{i}^{l}, h_{\mathcal{N}(i)}^{l+1} + b) \right)
        h_{i}^{(l+1)} & = \mathrm{norm}(h_{i}^{l})
    
    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied on a unidirectional bipartite graph, ``in_feats``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
        If aggregator type is ``gcn``, the feature size of source and destination nodes
        are required to be the same.
    out_size : int
        Output feature size.
    feat_drop : float
        Dropout rate on features, default: ``0``.
    aggregator_type : str
        Aggregator type to use (``mean``, ``gcn``, ``pool``, ``lstm``).
    bias : bool
        If True, adds a learnable bias to the output. Default: ``True``.
    norm : callable activation function/layer or None, optional
        If not None, applies normalization to the updated node features.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """
    def __init__(self,
                 in_feats,
                 out_feats,
                 aggregator_type,
                 feat_drop=0.,
                 bias=True,
                 norm=None,
                 activation=None):
        super(UndirectedGraphSAGELayerConv, self).__init__()

        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats
        self._aggre_type = aggregator_type
        self.norm = norm
        self.feat_drop = nn.Dropout(feat_drop)
        self.activation = activation
        # aggregator type: mean/pool/lstm/gcn
        if aggregator_type == 'pool':
            self.fc_pool = nn.Linear(self._in_src_feats, self._in_src_feats)
        if aggregator_type == 'lstm':
            self.lstm = nn.LSTM(self._in_src_feats, self._in_src_feats, batch_first=True)
        if aggregator_type != 'gcn':
            self.fc_self = nn.Linear(self._in_dst_feats, out_feats, bias=bias)
        self.fc_neigh = nn.Linear(self._in_src_feats, out_feats, bias=bias)
        self.reset_parameters()

    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain = nn.init.calculate_gain('relu')
        if self._aggre_type == 'pool':
            nn.init.xavier_uniform_(self.fc_pool.weight, gain=gain)
        if self._aggre_type == 'lstm':
            self.lstm.reset_parameters()
        if self._aggre_type != 'gcn':
            nn.init.xavier_uniform_(self.fc_self.weight, gain=gain)
        nn.init.xavier_uniform_(self.fc_neigh.weight, gain=gain)

    def _lstm_reducer(self, nodes):
        """LSTM reducer
        NOTE(zihao): lstm reducer with default schedule (degree bucketing)
        is slow, we could accelerate this with degree padding in the future.
        """
        m = nodes.mailbox['m'] # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_feats)),
             m.new_zeros((1, batch_size, self._in_src_feats)))
        _, (rst, _) = self.lstm(m, h)
        return {'neigh': rst.squeeze(0)}

    def forward(self, graph, feat, edge_weight=None,reverse_edge_weight=None):
        r"""Compute GraphSAGE layer.
        
        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat : torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.
        edge_weighht: torch.tensor
               Only needed when consider the edge weights in message passing.
        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, D_{out})` where :math:`D_{out}`
            is size of output feature.
        """
        graph = graph.local_var()

        if isinstance(feat, tuple):
            feat_src = self.feat_drop(feat[0])
            feat_dst = self.feat_drop(feat[1])
        else:
            feat_src = feat_dst = self.feat_drop(feat)

        h_self = feat_dst

        if self._aggre_type == 'mean':
            graph.srcdata['h'] = feat_src
            if edge_weight is None:
               graph.update_all(fn.copy_src('h', 'm'), fn.mean('m', 'neigh'))
            else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.mean('m', 'neigh'))
            h_neigh = graph.dstdata['neigh']
        elif self._aggre_type == 'gcn':
            check_eq_shape(feat)
            graph.srcdata['h'] = feat_src
            graph.dstdata['h'] = feat_dst     # same as above if homogeneous
            if edge_weight is None:
               graph.update_all(fn.copy_src('h', 'm'), fn.sum('m', 'neigh'))
            else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.sum('m', 'neigh'))
            # divide in_degrees
            degs = graph.in_degrees().to(feat_dst)
            h_neigh = (graph.dstdata['neigh'] + graph.dstdata['h']) / (degs.unsqueeze(-1) + 1)
        elif self._aggre_type == 'pool':
            graph.srcdata['h'] = F.relu(self.fc_pool(feat_src))
            if edge_weight is None:
               graph.update_all(fn.copy_src('h', 'm'), fn.max('m', 'neigh'))
            else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.max('m', 'neigh'))
            h_neigh = graph.dstdata['neigh']
        elif self._aggre_type == 'lstm':
            graph.srcdata['h'] = feat_src
            if edge_weight is None:
                graph.update_all(fn.copy_src('h', 'm'), self._lstm_reducer)
            else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), self._lstm_reducer)
            h_neigh = graph.dstdata['neigh']
        else:
            raise KeyError('Aggregator type {} not recognized.'.format(self._aggre_type))

        # GraphSAGE GCN does not require fc_self.
        if self._aggre_type == 'gcn':
            rst = self.fc_neigh(h_neigh)
        else:
            rst = self.fc_self(h_self) + self.fc_neigh(h_neigh)
        # activation
        if self.activation is not None:
            rst = self.activation(rst)
        # normalization
        if self.norm is not None:
            rst = self.norm(rst)
        return rst



class BiSepGraphSAGELayerConv(GNNLayerBase):
    r"""Bidirection version GraphSAGE layer from paper `Inductive Representation Learning on
    Large Graphs <https://arxiv.org/pdf/1706.02216.pdf>`__.
    
    .. math::
        h_{\mathcal{N}(i)}^{(l+1)} & = \mathrm{aggregate}
        \left(\{h_{j}^{l}, \forall j \in \mathcal{N}(i) \}\right)
        h_{i}^{(l+1)} & = \sigma \left(W \cdot \mathrm{concat}
        (h_{i}^{l}, h_{\mathcal{N}(i)}^{l+1} + b) \right)
        h_{i}^{(l+1)} & = \mathrm{norm}(h_{i}^{l})
   
    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied on a unidirectional bipartite graph, ``in_feats``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
        If aggregator type is ``gcn``, the feature size of source and destination nodes
        are required to be the same.
    output_size : int
        Output feature size.
    feat_drop : float
        Dropout rate on features, default: ``0``.
    aggregator_type : str
        Aggregator type to use (``mean``, ``gcn``, ``pool``, ``lstm``).
    bias : bool
        If True, adds a learnable bias to the output. Default: ``True``.
    norm : callable activation function/layer or None, optional
        If not None, applies normalization to the updated node features.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """

    def __init__(self, input_size,output_size,aggregator_type,
                 feat_drop=0.,
                 bias=True,
                 norm=None,
                 activation=None):
        super(BiSepGraphSAGELayerConv, self).__init__()

        self._in_src_size, self._in_dst_size = expand_as_pair(input_size)
        self._output_size = output_size
        self._aggre_type = aggregator_type
        self.norm = norm
        self.feat_drop = nn.Dropout(feat_drop)
        self.activation = activation
        # aggregator type: mean/pool/lstm/gcn
        if aggregator_type == 'pool': #aggreagator different for two direction's network
                self.fc_pool_fw = nn.Linear(self._in_src_size, self._in_src_size)
                self.fc_pool_bw = nn.Linear(self._in_src_size, self._in_src_size)
        if aggregator_type == 'lstm':
                self.lstm_fw = nn.LSTM(self._in_src_size, self._in_src_size, batch_first=True)
                self.lstm_bw = nn.LSTM(self._in_src_size, self._in_src_size, batch_first=True)
        if aggregator_type != 'gcn':
                self.fc_self_fw = nn.Linear(self._in_dst_size, output_size, bias=bias)
                self.fc_self_bw = nn.Linear(self._in_dst_size, output_size, bias=bias)

        self.fc_neigh_fw = nn.Linear(self._in_src_size, output_size, bias=bias)
        self.fc_neigh_bw = nn.Linear(self._in_src_size, output_size, bias=bias)

        self.reset_parameters()


    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain_fw = nn.init.calculate_gain('relu')
        gain_bw = nn.init.calculate_gain('relu')
        if self._aggre_type == 'pool':
            nn.init.xavier_uniform_(self.fc_pool_fw.weight, gain=gain_fw)
            nn.init.xavier_uniform_(self.fc_pool_bw.weight, gain=gain_bw)
        if self._aggre_type == 'lstm':
            self.lstm_fw.reset_parameters()
            self.lstm_bw.reset_parameters()
        if self._aggre_type != 'gcn':
            nn.init.xavier_uniform_(self.fc_self_fw.weight, gain=gain_fw)
            nn.init.xavier_uniform_(self.fc_neigh_bw.weight, gain=gain_bw)

    def _lstm_reducer_fw(self, nodes):
        """LSTM reducer
        NOTE(zihao): lstm reducer with default schedule (degree bucketing)
        is slow, we could accelerate this with degree padding in the future.
        """
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_size)),
             m.new_zeros((1, batch_size, self._in_src_size)))
        _, (rst, _) = self.lstm_fw(m, h)
        return {'neigh': rst.squeeze(0)}

    def _lstm_reducer_bw(self, nodes):
        """LSTM reducer
        NOTE(zihao): lstm reducer with default schedule (degree bucketing)
        is slow, we could accelerate this with degree padding in the future.
        """
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_size)),
             m.new_zeros((1, batch_size, self._in_src_size)))
        _, (rst, _) = self.lstm_bw(m, h)
        return {'neigh': rst.squeeze(0)}

    def message_reduce(self,graph,direction,feat,edge_weight):
       if isinstance(feat, tuple):
        feat_src = self.feat_drop(feat[0])
        feat_dst = self.feat_drop(feat[1])
       else:
        feat_src = feat_dst = self.feat_drop(feat)

       h_self = feat_dst

       if self._aggre_type == 'mean':
         graph.srcdata['h'] = feat_src
         if edge_weight is None:
             graph.update_all(fn.copy_src('h', 'm'), fn.mean('m', 'neigh'))
         else:
             graph.edata['edge_weight']=edge_weight
             graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.sum('m', 'neigh'))
         h_neigh = graph.dstdata['neigh']
       elif self._aggre_type == 'gcn':
         check_eq_shape(feat)
         graph.srcdata['h'] = feat_src
         graph.dstdata['h'] = feat_dst  # same as above if homogeneous
         if edge_weight is None:
             graph.update_all(fn.copy_src('h', 'm'), fn.sum('m', 'neigh'))
             # divide in_degrees
             degs = graph.in_degrees().to(feat_dst)
             h_neigh = (graph.dstdata['neigh'] + graph.dstdata['h']) / (degs.unsqueeze(-1) + 1)             
         else:
             graph.edata['edge_weight']=edge_weight
             graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.sum('m', 'neigh'))

       elif self._aggre_type == 'pool':
         if direction=='fw':
             graph.srcdata['h'] = F.relu(self.fc_pool_fw(feat_src))
         elif direction=='bw':
             graph.srcdata['h'] = F.relu(self.fc_pool_bw(feat_src))
         if edge_weight is None:
             graph.update_all(fn.copy_src('h', 'm'), fn.max('m', 'neigh'))
         else:
             graph.edata['edge_weight']=edge_weight
             graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.max('m', 'neigh'))
         h_neigh = graph.dstdata['neigh']
       elif self._aggre_type == 'lstm':
         graph.srcdata['h'] = feat_src
         if direction=='fw':
           if edge_weight is None:
               graph.update_all(fn.copy_src('h', 'm'), self._lstm_reducer_fw)
           else:
              warnings.warn('Operating lstm aggregation by using the edge weights in GNN', SyntaxWarning)
              graph.edata['edge_weight']=edge_weight
              graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), self._lstm_reducer_fw)
         elif direction=='bw':
           if edge_weight is None:
              graph.update_all(fn.copy_src('h', 'm'), self._lstm_reducer_bw)
           else:
              warnings.warn('Operating lstm aggregation by using the edge weights in GNN', SyntaxWarning)
              graph.edata['edge_weight']=edge_weight
              graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), self._lstm_reducer_bw)

         h_neigh = graph.dstdata['neigh']
       else:
        raise KeyError('Aggregator type {} not recognized.'.format(self._aggre_type))


       return h_neigh,h_self

    def forward(self,graph,feat,edge_weight=None,reverse_edge_weight=None):
        r"""      
        Compute node embeddings from both directions in bidirection seperated GraphSAGE
        
        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat_fw : torch.Tensor or pair of torch.Tensor
        feat_bw: torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.
        edge_weighht: torch.tensor
               Only needed when consider the edge weights in message passing.
       
        Returns
        -------
        The output feature of shape :math:`(N, D_{out})` where :math:`D_{out}`
        is size of output feature.
        """
        if isinstance(feat,list):   #judge whether the the input is the initial node feature or the two outputs from the last BisepGraphSAGELayer
           feat_fw, feat_bw = feat
        else:
           feat_fw = feat
           feat_bw = feat

        self.forward_graph = graph
        self.backward_graph = graph.reverse()
        f_graph = self.forward_graph.local_var()
        b_graph = self.forward_graph.local_var()



        # update node part:
        h_neigh_fw,h_self_fw = self.message_reduce(f_graph,'fw',feat_fw,edge_weight)
        h_neigh_bw,h_self_bw = self.message_reduce(b_graph,'bw',feat_bw,reverse_edge_weight)
        # GraphSAGE GCN does not require fc_self.


        if self._aggre_type == 'gcn':
            rst_fw = self.fc_neigh_fw(h_neigh_fw)
            rst_bw = self.fc_neigh_bw(h_neigh_bw)
        else:
            rst_fw = self.fc_self_fw(h_self_fw) + self.fc_neigh_fw(h_neigh_fw)
            rst_bw = self.fc_self_bw(h_self_bw) + self.fc_neigh_bw(h_neigh_bw)

        # activation
        if self.activation is not None:
            rst_fw = self.activation(rst_fw)
            rst_bw = self.activation(rst_bw)
            # normalization
        if self.norm is not None:
            rst_fw = self.norm(rst_fw)
            rst_bw = self.norm(rst_bw)

        return [rst_fw,rst_bw]


class BiFuseGraphSAGELayerConv(GNNLayerBase):
    r"""Bidirection version GraphSAGE layer from paper `Inductive Representation Learning on
    Large Graphs <https://arxiv.org/pdf/1706.02216.pdf>`__.
    
    .. math::
        h_{\mathcal{N}(i)}^{(l+1)} & = \mathrm{aggregate}
        \left(\{h_{j}^{l}, \forall j \in \mathcal{N}(i) \}\right)
        h_{i}^{(l+1)} & = \sigma \left(W \cdot \mathrm{concat}
        (h_{i}^{l}, h_{\mathcal{N}(i)}^{l+1} + b) \right)
        h_{i}^{(l+1)} & = \mathrm{norm}(h_{i}^{l})
    
    Parameters
    ----------
    input_size : int, or pair of ints
        Input feature size.
        If the layer is to be applied on a unidirectional bipartite graph, ``in_feats``
        specifies the input feature size on both the source and destination nodes.  If
        a scalar is given, the source and destination node feature size would take the
        same value.
        If aggregator type is ``gcn``, the feature size of source and destination nodes
        are required to be the same.
    output_size : int
        Output feature size.
    feat_drop : float
        Dropout rate on features, default: ``0``.
    aggregator_type : str
        Aggregator type to use (``mean``, ``gcn``, ``pool``, ``lstm``).
    bias : bool
        If True, adds a learnable bias to the output. Default: ``True``.
    norm : callable activation function/layer or None, optional
        If not None, applies normalization to the updated node features.
    activation : callable activation function/layer or None, optional
        If not None, applies an activation function to the updated node features.
        Default: ``None``.
    """

    def __init__(self, input_size,output_size,aggregator_type,
                 feat_drop=0.,
                 bias=True,
                 norm=None,
                 activation=None):
        super(BiFuseGraphSAGELayerConv, self).__init__()

        self._in_src_size, self._in_dst_size = expand_as_pair(input_size)
        self._output_size = output_size
        self._aggre_type = aggregator_type
        self.norm = norm
        self.feat_drop = nn.Dropout(feat_drop)
        self.activation = activation
        # aggregator type: mean/pool/lstm/gcn
        if aggregator_type == 'pool': #aggreagator different for two direction's network
                self.fc_pool_fw = nn.Linear(self._in_src_size, self._in_src_size)
                self.fc_pool_bw = nn.Linear(self._in_src_size, self._in_src_size)
        if aggregator_type == 'lstm':
                self.lstm_fw = nn.LSTM(self._in_src_size, self._in_src_size, batch_first=True)
                self.lstm_bw = nn.LSTM(self._in_src_size, self._in_src_size, batch_first=True)
        if aggregator_type != 'gcn':
                self.fc_self_fw = nn.Linear(self._in_dst_size, output_size, bias=bias)
                self.fc_self_bw = nn.Linear(self._in_dst_size, output_size, bias=bias)

        self.fc_neigh = nn.Linear(self._in_src_size, output_size, bias=bias)

        self.fuse_linear=nn.Linear(4*self._in_src_size, self._in_src_size, bias=bias)

        self.reset_parameters()


    def reset_parameters(self):
        """Reinitialize learnable parameters."""
        gain_fw = nn.init.calculate_gain('relu')
        gain_bw = nn.init.calculate_gain('relu')
        if self._aggre_type == 'pool':
            nn.init.xavier_uniform_(self.fc_pool_fw.weight, gain=gain_fw)
            nn.init.xavier_uniform_(self.fc_pool_bw.weight, gain=gain_bw)
        if self._aggre_type == 'lstm':
            self.lstm_fw.reset_parameters()
            self.lstm_bw.reset_parameters()
        if self._aggre_type != 'gcn':
            nn.init.xavier_uniform_(self.fc_self_fw.weight, gain=gain_fw)
            nn.init.xavier_uniform_(self.fc_self_bw.weight, gain=gain_bw)

    def _lstm_reducer_fw(self, nodes):
        """LSTM reducer
        NOTE(zihao): lstm reducer with default schedule (degree bucketing)
        is slow, we could accelerate this with degree padding in the future.
        """
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_size)),
             m.new_zeros((1, batch_size, self._in_src_size)))
        _, (rst, _) = self.lstm_fw(m, h)
        return {'neigh': rst.squeeze(0)}

    def _lstm_reducer_bw(self, nodes):
        """LSTM reducer
        NOTE(zihao): lstm reducer with default schedule (degree bucketing)
        is slow, we could accelerate this with degree padding in the future.
        """
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_size)),
             m.new_zeros((1, batch_size, self._in_src_size)))
        _, (rst, _) = self.lstm_bw(m, h)
        return {'neigh': rst.squeeze(0)}

    def forward(self,graph,feat,edge_weight=None, reverse_edge_weight=None):
        r"""
        Compute node embeddings from both directions in bidirection seperated GraphSAGE
        
        Parameters
        ----------
        graph : DGLGraph
            The graph.
        feat_fw : torch.Tensor or pair of torch.Tensor
        feat_bw: torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.
        edge_weighht: torch.tensor
               Only needed when consider the edge weights in message passing.
        
        Returns
        -------
        The output feature of shape :math:`(N, D_{out})` where :math:`D_{out}`
        is size of output feature.
        """
        if feat is list:   #judge whether the the input is the initial node feature or the two outputs from the last BisepGraphSAGELayer
           feat_fw, feat_bw = feat
        else:
           feat_fw = feat
           feat_bw = feat

        self.forward_graph = graph
        self.backward_graph = graph.reverse()
        f_graph = self.forward_graph.local_var()
        b_graph = self.forward_graph.local_var()

        def fuse(self,forward_message,backward_message):
            cat=torch.cat([forward_message,backward_message],dim=1)
            sum=forward_message+backward_message
            diff=forward_message-backward_message
            if self.activation:
              z=self.activation(self.fuse_linear(torch.cat([cat,sum,diff],dim=1)))
            else:
              z=self.fuse_linear(torch.cat([cat,sum,diff],dim=1))

            return z*forward_message+(1-z)*backward_message

        def message_reduce(self,graph,feat,direction,edge_weight):
           if isinstance(feat, tuple):
            feat_src = self.feat_drop(feat[0])
            feat_dst = self.feat_drop(feat[1])
           else:
            feat_src = feat_dst = self.feat_drop(feat)

           h_self = feat_dst

           if self._aggre_type == 'mean':
             graph.srcdata['h'] = feat_src
             if edge_weight is None:
               graph.update_all(fn.copy_src('h', 'm'), fn.mean('m', 'neigh'))
             else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.sum('m', 'neigh'))
             h_neigh = graph.dstdata['neigh']
           elif self._aggre_type == 'gcn':
             check_eq_shape(feat)
             graph.srcdata['h'] = feat_src
             graph.dstdata['h'] = feat_dst  # same as above if homogeneous
             if edge_weight is None:
                 graph.update_all(fn.copy_src('h', 'm'), fn.sum('m', 'neigh'))
                 # divide in_degrees
                 degs = graph.in_degrees().to(feat_dst)
                 h_neigh = (graph.dstdata['neigh'] + graph.dstdata['h']) / (degs.unsqueeze(-1) + 1)                      
             else:
                 graph.edata['edge_weight']=edge_weight
                 graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.sum('m', 'neigh'))
           elif self._aggre_type == 'pool':
             if direction=='fw':
                 graph.srcdata['h'] = F.relu(self.fc_pool_fw(feat_src))
             elif direction=='bw':
                 graph.srcdata['h'] = F.relu(self.fc_pool_bw(feat_src))
             if edge_weight is None:
                graph.update_all(fn.copy_src('h', 'm'), fn.max('m', 'neigh'))
             else:
               graph.edata['edge_weight']=edge_weight
               graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), fn.max('m', 'neigh'))
             h_neigh = graph.dstdata['neigh']
           elif self._aggre_type == 'lstm':
             graph.srcdata['h'] = feat_src
             if direction=='fw':
               if edge_weight is None:
                   graph.update_all(fn.copy_src('h', 'm'), self._lstm_reducer_fw)
               else:
                   warnings.warn('Operating lstm aggregation by using the edge weights in GNN', SyntaxWarning)
                   graph.edata['edge_weight']=edge_weight
                   graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), self._lstm_reducer_fw)
             elif direction=='bw':
               if edge_weight is None:
                   graph.update_all(fn.copy_src('h', 'm'), self._lstm_reducer_bw)
               else:
                   warnings.warn('Operating lstm aggregation by using the edge weights in GNN', SyntaxWarning)                  
                   graph.edata['edge_weight']=edge_weight
                   graph.update_all(fn.u_mul_e('h', 'edge_weight','m'), self._lstm_reducer_bw)

             h_neigh = graph.dstdata['neigh']
           else:
            raise KeyError('Aggregator type {} not recognized.'.format(self._aggre_type))

           return h_neigh,h_self

        # update node part:
        h_neigh_fw,h_self_fw = message_reduce(self,f_graph,feat_fw,'fw',edge_weight)
        h_neigh_bw,h_self_bw = message_reduce(self, b_graph,feat_bw,'bw',reverse_edge_weight)

        #fuse the two directions' information
        h_neigh_fused=fuse(self,h_neigh_fw,h_neigh_bw)
        # GraphSAGE GCN does not require fc_self.
        if self._aggre_type == 'gcn':
            rst_fused = self.fc_neigh(h_neigh_fused)
        else:
            rst_fused = self.fc_self_fw(h_self_fw) + self.fc_neigh(h_neigh_fused)

        # activation
        if self.activation is not None:
            rst_fused = self.activation(rst_fused)
            # normalization
        if self.norm is not None:
            rst_fused = self.norm(rst_fused)

        return rst_fused



#!/usr/bin/env python

# Python wrapper for METEOR implementation, by Xinlei Chen
# Acknowledge Michael Denkowski for the generous discussion and help

import os
import sys
import subprocess
import threading

# Assumes meteor_tool-1.5.jar is in the same directory as meteor_tool.py.  Change as needed.
METEOR_JAR = 'meteor-1.5.jar'
# print METEOR_JAR

class Meteor:

    def __init__(self):
        self.meteor_cmd = ['java', '-jar', '-Xmx2G', METEOR_JAR,
                '-', '-', '-stdio', '-l', 'en', '-norm']
        self.meteor_p = subprocess.Popen(self.meteor_cmd,
                                         cwd=os.path.dirname(os.path.abspath(__file__)),
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
        # self.meteor_p = subprocess.Popen(self.meteor_cmd, \
        #         cwd=os.path.dirname(os.path.abspath(__file__)), \
        #         stdin=subprocess.PIPE, \
        #         stdout=subprocess.PIPE, \
        #         stderr=subprocess.PIPE)
        # Used to guarantee thread safety
        self.lock = threading.Lock()

    def compute_score(self, gts, res):
        assert(gts.keys() == res.keys())
        imgIds = gts.keys()
        scores = []

        eval_line = 'EVAL'
        self.lock.acquire()
        for i in imgIds:
            assert(len(res[i]) == 1)
            stat = self._stat(res[i][0], gts[i])
            eval_line += ' ||| {}'.format(stat)

        self.meteor_p.stdin.write('{}\n'.format(eval_line).encode())
        self.meteor_p.stdin.flush()
        for i in range(0,len(imgIds)):
            scores.append(float(self.meteor_p.stdout.readline().strip()))
        score = float(self.meteor_p.stdout.readline().strip())
        self.lock.release()

        return score, scores

    def method(self):
        return "METEOR"

    def _stat(self, hypothesis_str, reference_list):
        # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
        hypothesis_str = hypothesis_str.replace('|||','').replace('  ',' ')
        score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
        self.meteor_p.stdin.write('{}\n'.format(score_line).encode())
        self.meteor_p.stdin.flush()
        return self.meteor_p.stdout.readline().decode().strip()

    def _score(self, hypothesis_str, reference_list):
        self.lock.acquire()
        # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
        hypothesis_str = hypothesis_str.replace('|||','').replace('  ',' ')
        score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
        self.meteor_p.stdin.write('{}\n'.format(score_line))
        stats = self.meteor_p.stdout.readline().strip()
        eval_line = 'EVAL ||| {}'.format(stats)
        # EVAL ||| stats
        self.meteor_p.stdin.write('{}\n'.format(eval_line))
        score = float(self.meteor_p.stdout.readline().strip())
        # bug fix: there are two values returned by the jar file, one average, and one all, so do it twice
        # thanks for Andrej for pointing this out
        score = float(self.meteor_p.stdout.readline().strip())
        self.lock.release()
        return score

    def __del__(self):
        self.lock.acquire()
        self.meteor_p.stdin.close()
        self.meteor_p.kill()
        self.meteor_p.wait()
        self.lock.release()



import copy
import json

import torch
from pythonds.basic.stack import Stack
from stanfordcorenlp import StanfordCoreNLP

from graph4nlp.pytorch.data.data import GraphData, to_batch
from .base import StaticGraphConstructionBase
from .utils import CORENLP_TIMEOUT_SIGNATURE


class ConstituencyBasedGraphConstruction(StaticGraphConstructionBase):
    """
    Class for constituency graph construction.

    ...

    Attributes
    ----------
    embedding_styles : (dict)
        Specify embedding styles including ``single_token_item``, ``emb_strategy``, ``num_rnn_layers``, ``bert_model_name`` and ``bert_lower_case``.

    vocab: (set, optional)
        Vocabulary including all words appeared in graphs.

    Methods
    -------

    topology(raw_text_data, nlp_processor, merge_strategy=None, edge_strategy=None)
        Generate graph structure with nlp parser like ``CoreNLP`` etc.

    _construct_static_graph(parsed_object, sub_sentence_id, edge_strategy=None)
        Construct a single static graph from a single sentence, to be called by ``topology`` function.

    _graph_connect(nx_graph_list, merge_strategy=None)
        Construct a merged graph from a list of graphs, to be called by ``topology`` function.

    embedding(node_attributes, edge_attributes)
        Generate node/edge embeddings from node/edge attributes through an embedding layer.

    forward(raw_text_data, nlp_parser)
        Generate graph topology and embeddings.
    """

    def __init__(self, embedding_style, 
                       vocab, 
                       hidden_size, 
                       fix_word_emb=True, 
                       fix_bert_emb=True, 
                       word_dropout=None,
                       rnn_dropout=None):
        super(ConstituencyBasedGraphConstruction, self).__init__(word_vocab=vocab,
                                                                 embedding_styles=embedding_style,
                                                                 hidden_size=hidden_size,
                                                                 fix_word_emb=fix_word_emb,
                                                                 fix_bert_emb=fix_bert_emb,
                                                                 word_dropout=word_dropout,
                                                                 rnn_dropout=rnn_dropout)
        self.vocab = vocab

    @classmethod
    def parsing(cls, raw_text_data, 
                     nlp_processor, 
                     processor_args):
        '''
        Parameters
        ----------
        raw_text_data: str
        nlp_processor: StanfordCoreNLP
        processor_args: json config for constituency graph construction
        '''
        output = nlp_processor.annotate(raw_text_data.strip().replace('(', '<LB>').replace(')', '<RB>'), properties=processor_args)
        if CORENLP_TIMEOUT_SIGNATURE in output:
            raise TimeoutError('CoreNLP timed out at input: \n{}\n This item will be skipped. '
                               'Please check the input or change the timeout threshold.'.format(raw_text_data))
        parsed_output = json.loads(output)['sentences']
        return parsed_output

    @classmethod
    def topology(cls,
                 raw_text_data,
                 nlp_processor,
                 processor_args,
                 merge_strategy=None,
                 edge_strategy=None,
                 verbase=True):
        """topology This function generate a graph strcuture from a raw text data.

        Parameters
        ----------
        raw_text_data : string
            A string to be used to construct a static graph, can be composed of multiple strings

        nlp_processor : object
            A parser used to parse sentence string to parsing trees like dependency parsing tree or constituency parsing tree

        merge_strategy : None or str, option=[None, "tailhead", "user_define"]
            Strategy to merge sub-graphs into one graph
            ``None``: It will be the default option. We will do as ``"tailhead"``.
            ``"tailhead"``: Link the sub-graph  ``i``'s tail node with ``i+1``'s head node
            ``"user_define"``: We will give this option to the user. User can override the method ``_graph_connnect`` to define your merge
                               strategy.

        edge_strategy: None or str, option=[None, "homogeneous", "heterogeneous", "as_node"]
            Strategy to process edge.
            ``None``: It will be the default option. We will do as ``"homogeneous"``.
            ``"homogeneous"``: We will drop the edge type information.
                               If there is a linkage among node ``i`` and node ``j``, we will add an edge whose weight
                               is ``1.0``. Otherwise there is no edge.
            ``heterogeneous``: We will keep the edge type information.
                               An edge will have type information like ``n_subj``.
                               It is not implemented yet.
            ``as_node``: We will view the edge as a graph node.
                         If there is an edge whose type is ``k`` between node ``i`` and node ``j``,
                         we will insert a node ``k`` into the graph and link node (``i``, ``k``) and (``k``, ``j``).
                         It is not implemented yet.

        verbase : bool
            A boolean option to decide whether to print out the graph construction process.

        Returns
        -------
        GraphData
            A customized graph data structure
        """
        output_graph_list = []
        parsed_output = cls.parsing(raw_text_data, nlp_processor, processor_args)
        for index in range(len(parsed_output)):
            # print(parsed_output[index]['parse'])
            output_graph_list.append(
                cls._construct_static_graph(parsed_output[index], index))
        ret_graph = cls._graph_connect(output_graph_list)
        # for edge_item in ret_graph.get_all_edges():
        #     print(ret_graph.node_attributes[edge_item[0]]['token'], "->", ret_graph.node_attributes[edge_item[1]]['token'])
        if verbase:
            print('--------------------------------------')
            for _edge in ret_graph.get_all_edges():
                print(ret_graph.nodes[_edge[0]].attributes['token'],
                      "\t---\t", ret_graph.nodes[_edge[1]].attributes['token'])
            print('--------------------------------------')

        return ret_graph

    @classmethod
    def _construct_static_graph(cls,
                                parsed_object,
                                sub_sentence_id,
                                edge_strategy=None,
                                sequential_link=False,
                                bisequential_link=True,
                                top_down=False,
                                add_pos_node=False,
                                cut_line_node=True):
        """construct single syntactic graph

        Parameters
        ----------
        parsed_object :
            A parsed object of single sentence after external parser like ``CoreNLP``

        sub_sentence_id : int
            To specify which sentence the ``parsed_object`` is in the ``raw_text_data``

        bisequential_link : bool
            Add bi-directional edges between word nodes, do not add if ``False``

        top_down : bool
            Edge direction between nodes in constituency tree, if ``True``, add edges from top to down

        add_pos_node : bool
            Add part-of-speech nodes or not

        cut_line_node : bool
            Cut nodes which are organized in "line" format (degree = 2)
        """
        assert (add_pos_node & cut_line_node) == False
        parsed_sentence_data = parsed_object['parse']
        for punc in [u'(', u')']:
            parsed_sentence_data = parsed_sentence_data.replace(
                punc, ' ' + punc + ' ')
        parse_list = (parsed_sentence_data.strip()).split()
        # cut root node
        if parse_list[0] == "(" and parse_list[1] == "ROOT":
            parse_list = parse_list[2:-1]
        # transform '.' to 'period'
        for index in range(len(parse_list)):
            if index <= len(parse_list) - 2:
                if parse_list[index] == "." and parse_list[index+1] == ".":
                    parse_list[index] = "period"
        res_graph = GraphData()
        pstack = Stack()
        idx = 0
        cnt_word_node = -1
        while idx < len(parse_list):
            if parse_list[idx] == '(':
                res_graph.add_nodes(1)
                # res_graph.node_attributes[res_graph.get_node_num(
                # ) - 1] = {'token': parse_list[idx + 1], 'type': 1, 'position_id': None, 'sentence_id': sub_sentence_id,
                #           'tail': False, 'head': False}
                res_graph.node_attributes[res_graph.get_node_num() - 1] = {'token': parse_list[idx + 1], 'type': 1, 'position_id': None, 'sentence_id': sub_sentence_id,
                          'tail': False, 'head': False}
                pstack.push(res_graph.get_node_num() - 1)
                if pstack.size() > 1:
                    node_2 = pstack.pop()
                    node_1 = pstack.pop()
                    if top_down:
                        res_graph.add_edge(node_1, node_2)
                    else:
                        res_graph.add_edge(node_2, node_1)
                    pstack.push(node_1)
                    pstack.push(node_2)
            elif parse_list[idx] == ')':
                pstack.pop()
            elif parse_list[idx + 1] == u')' and parse_list[idx] != u')':
                cnt_word_node += 1
                if add_pos_node:
                    res_graph.add_nodes(1)
                else:
                    assert res_graph.node_attributes[res_graph.get_node_num()-1]['type'] != 0
                res_graph.node_attributes[res_graph.get_node_num(
                ) - 1] = {'token': parse_list[idx], 'type': 0, 'position_id': cnt_word_node,
                          'sentence_id': sub_sentence_id, 'tail': False, 'head': False}
                node_1 = pstack.pop()
                if add_pos_node:
                    assert node_1 != res_graph.get_node_num() - 1
                    if top_down:
                        res_graph.add_edge(node_1, res_graph.get_node_num() - 1)
                    else:
                        res_graph.add_edge(res_graph.get_node_num() - 1, node_1)
                else:
                    assert node_1 == res_graph.get_node_num() - 1
                pstack.push(node_1)
            idx += 1

        if sequential_link:
            _len_single_graph = res_graph.get_node_num()
            _cnt_node = 0
            while (True):
                node_1_idx = -1
                node_2_idx = -1
                for idx_seq, node_attr in enumerate(res_graph.node_attributes):
                    if node_attr['position_id'] == _cnt_node:
                        node_1_idx = idx_seq
                    elif node_attr['position_id'] == _cnt_node + 1:
                        node_2_idx = idx_seq
                if node_1_idx != -1 and node_2_idx != -1:
                    res_graph.add_edge(node_1_idx, node_2_idx)
                    if bisequential_link:
                        res_graph.add_edge(node_2_idx, node_1_idx)
                if _cnt_node >= _len_single_graph:
                    break
                _cnt_node += 1

        max_pos = 0
        for n in res_graph.node_attributes:
            if n['type'] == 0 and n['position_id'] > max_pos:
                max_pos = n['position_id']

        min_pos = 99999
        for n in res_graph.node_attributes:
            if n['type'] == 0 and n['position_id'] < min_pos:
                min_pos = n['position_id']

        for n in res_graph.node_attributes:
            if n['type'] == 0 and n['position_id'] == max_pos:
                n['tail'] = True
            if n['type'] == 0 and n['position_id'] == min_pos:
                n['head'] = True
        # print(res_graph.edges)
        if cut_line_node:
            res_graph = cls._cut_line_node(res_graph)
        return res_graph

    @classmethod
    def _graph_connect(cls, graph_list, merge_strategy=None, bisequential_link=True, reformalize=True):
        """
        Parameters
        ----------
        graph_list : list
            A graph list to be merged

        bisequential_link : bool
            whether add bi-direnctional links between word nodes

        reformalize : bool
            If true, separate word nodes and non-terminal nodes in ``graph.node_attributes`` and put word nodes in the front position

        bisequential_link : bool
            whether add bi-direnctional links between word nodes

        reformalize : bool
            If true, separate word nodes and non-terminal nodes in ``graph.node_attributes`` and put word nodes in the front position

        Returns
        -------
        GraphData
            A customized graph data structure
        """
        _len_graph_ = len(graph_list)
        # if only one sentence included, we do not need to merge the graph.
        if _len_graph_ > 1:
            merged_graph = GraphData()
            for index in range(_len_graph_):
                len_merged_graph = merged_graph.get_node_num()

                graph_i_nodes_attributes = []
                for list_item in graph_list[index].node_attributes:
                    graph_i_nodes_attributes.append(list_item)
                merged_graph.node_attributes.extend(graph_i_nodes_attributes)

            _cnt_edge_num = 0
            for g_idx in range(_len_graph_):
                tmp_edges = graph_list[g_idx].get_all_edges()
                current_node_num = graph_list[g_idx].get_node_num()
                for _edge in tmp_edges:
                    merged_graph.add_edge(
                        _edge[0] + _cnt_edge_num, _edge[1] + _cnt_edge_num)
                _cnt_edge_num += current_node_num

            for index in range(_len_graph_ - 1):
                # get head node
                head_node = -1
                for idx, node_attr_item in enumerate(merged_graph.node_attributes):
                    if node_attr_item['sentence_id'] == index + 1 and node_attr_item['type'] == 0 and node_attr_item['head'] == True:
                        head_node = idx
                # get tail node
                tail_node = -1
                for idx, node_attr_item in enumerate(merged_graph.node_attributes):
                    if node_attr_item['sentence_id'] == index and node_attr_item['type'] == 0 and node_attr_item['tail'] == True:
                        tail_node = idx
                merged_graph.add_edge(tail_node, head_node)
                if bisequential_link:
                    merged_graph.add_edge(head_node, tail_node)
        else:
            merged_graph = graph_list[0]
        if reformalize:
            new_list_for_nodes = copy.deepcopy(merged_graph.node_attributes)

            list_for_word_nodes = []
            list_for_non_ternimal_nodes = []
            for i_0, i_1 in enumerate(new_list_for_nodes):
                if i_1['type'] == 0:
                    list_for_word_nodes.append((i_0, i_1))
                else:
                    list_for_non_ternimal_nodes.append((i_0, i_1))

            node_id_map = {}
            cnt = 0
            merged_graph.node_attributes.clear()
            for item in list_for_word_nodes:
                merged_graph.node_attributes.append(item[1])
                node_id_map[item[0]] = cnt
                cnt += 1
            for item in list_for_non_ternimal_nodes:
                merged_graph.node_attributes.append(item[1])
                node_id_map[item[0]] = cnt
                cnt += 1
            for i in range(merged_graph.get_edge_num()):
                merged_graph._edge_indices.src[i] = node_id_map[merged_graph._edge_indices.src[i]]
                merged_graph._edge_indices.tgt[i] = node_id_map[merged_graph._edge_indices.tgt[i]]
        return merged_graph

    @classmethod
    def _cut_line_node(cls, input_graph: GraphData):
        idx_to_be_deleted = []
        new_edges = []
        for idx, n in enumerate(input_graph.node_attributes):
            edge_arr = input_graph.get_all_edges()
            cnt_in = 0
            cnt_out = 0
            for e in edge_arr:
                if idx == e[0]:
                    cnt_out += 1
                    out_ = e[1]
                if idx == e[1]:
                    cnt_in += 1
                    in_ = e[0]
            if cnt_in == 1 and cnt_out == 1:
                idx_to_be_deleted.append(idx)
                new_edges.append((in_, out_))
        if len(idx_to_be_deleted) == 0:
            return input_graph
        res_graph = GraphData()
        id_map = {}
        cnt_node = 0
        for idx, n in enumerate(input_graph.node_attributes):
            if idx not in idx_to_be_deleted:
                res_graph.add_nodes(1)
                res_graph.node_attributes[res_graph.get_node_num()-1] = n
                id_map[idx] = cnt_node
                cnt_node += 1
            else:
                id_map[idx] = -1
        for edge_arr in input_graph.get_all_edges()+new_edges:
            if (edge_arr[0] not in idx_to_be_deleted) and (edge_arr[1] not in idx_to_be_deleted):
                res_graph.add_edge(id_map[edge_arr[0]], id_map[edge_arr[1]])
        return res_graph

    def forward(self, batch_graphdata: list):
        batch_graphdata = self.embedding_layer(batch_graphdata)
        return batch_graphdata



from .ComplEx import ComplEx
from .ComplExLayer import ComplExLayer
from .DistMult import DistMult
from .DistMultLayer import DistMultLayer

__all__ = ['ComplEx', 'ComplExLayer', 'DistMult', 'DistMultLayer']


from .ConcatFeedForwardNN import ConcatFeedForwardNN
from .ConcatFeedForwardNNLayer import ConcatFeedForwardNNLayer
from .ElementSum import ElementSum 
from .ElementSumLayer import ElementSumLayer
from .StackedElementProd import StackedElementProd
from .StackedElementProdLayer import StackedElementProdLayer

__all__ = ['ConcatFeedForwardNN', 'ConcatFeedForwardNNLayer', 'ElementSum', 'ElementSumLayer',
            'StackedElementProd', 'StackedElementProdLayer']


from ....utils.config_utils import get_yaml_config
import os

str2yaml = {"stdrnn": "stdrnndecoder.yaml", "stdtree": "stdtreedecoder.yaml"}
dir_path = os.path.dirname(os.path.realpath(__file__))


def get_decoder_args(deocder_name):
    """
        It will build the template for ``DecoderBase`` model.
    Parameters
    ----------
    decoder_name: str
        The decoder name. Expected in ["stdrnn", "stdtree"].

    Returns
    -------
    template_dict: dict
        The template dict.
        The structure is shown as follows:
        {
            rnn_decoder_share: {rnn_type: "lstm", input_size: 300, ...},
            rnn_decoder_private: {max_decoder_step: 50}
        }
        The ``rnn_decoder_share`` contains the parameters shared by all ``DecoderBase`` models.
        The ``rnn_decoder_private`` contains the parameters specifically in each decoder methods.
    """
    if deocder_name in str2yaml.keys():
        yaml_name = str2yaml[deocder_name]
        path = os.path.join(dir_path, yaml_name)
        config = get_yaml_config(path)
        return config
    else:
        return {}


__all__ = ["get_decoder_args"]



import copy

import torch.nn.functional as F
import torch

from graph4nlp.pytorch.data.data import from_batch
from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from .base import Graph2XBase
from graph4nlp.pytorch.modules.prediction.generation.decoder_strategy import DecoderStrategy
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel
from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import WordEmbedding



class Graph2Seq(Graph2XBase):
    """
        The graph2seq model consists the following components: 1) node embedding 2) graph embedding 3) decoding.
        Since the full pipeline will consist all parameters, so we will add prefix to the original parameters
         in each component as follows (except the listed four parameters):
            1) emb_ + parameter_name (eg: ``emb_input_size``)
            2) gnn_ + parameter_name (eg: ``gnn_direction_option``)
            3) dec_ + parameter_name (eg: ``dec_max_decoder_step``)
        Considering neatness, we will only present the four hyper-parameters which don't meet regulations.
        Examples
        -------
        # Build a vocab model from scratch
        >>> "It is just a how-to-use example."
        >>> from graph4nlp.pytorch.modules.config import get_basic_args
        >>> opt = get_basic_args(graph_construction_name="node_emb", graph_embedding_name="gat", decoder_name="stdrnn")
        >>> graph2seq = Graph2Seq.from_args(opt=opt, vocab_model=vocab_model, device=torch.device("cuda:0"))
        >>> batch_graph = [GraphData() for _ in range(2)]
        >>> tgt_seq = torch.Tensor([[1, 2, 3], [4, 5, 6]])
        >>> seq_out, _, _ = graph2seq(batch_graph=batch_graph, tgt_seq=tgt_seq)
        >>> print(seq_out.shape) # [2, 6, 5] (assume the vocabulary size is 5 and max_decoder_step is 6)
    Parameters
    ----------
    vocab_model: VocabModel
        The vocabulary.
    graph_type: str
        The graph type. Excepted in ["dependency", "constituency", "node_emb", "node_emb_refined"].
    gnn: str
        The graph neural network's name. Expected in ["gcn", "gat", "graphsage", "ggnn"]
    """
    def __init__(self, vocab_model, emb_input_size, emb_hidden_size, embedding_style,
                 graph_type, gnn_direction_option, gnn_input_size, gnn_hidden_size, gnn_output_size,
                 gnn, gnn_num_layers, dec_hidden_size, share_vocab=False,
                 # dropout
                 gnn_feat_drop=0.0, gnn_attn_drop=0.0,
                 emb_fix_word_emb=False, emb_fix_bert_emb=False, emb_word_dropout=0.0, emb_rnn_dropout=0.0,
                 dec_max_decoder_step=50,
                 dec_use_copy=False, dec_use_coverage=False,
                 dec_graph_pooling_strategy=None, dec_rnn_type="lstm", dec_tgt_emb_as_output_layer=False, dec_teacher_forcing_rate=1.0,
                 dec_attention_type="uniform", dec_fuse_strategy="average", dec_node_type_num=None,
                 dec_dropout=0.0,
                 **kwargs):
        super(Graph2Seq, self).__init__(vocab_model=vocab_model, emb_input_size=emb_input_size, emb_hidden_size=emb_hidden_size,
                                        graph_type=graph_type, gnn_direction_option=gnn_direction_option,
                                        gnn=gnn, gnn_num_layers=gnn_num_layers, embedding_style=embedding_style,
                                        gnn_feats_dropout=gnn_feat_drop,
                                        gnn_attn_dropout=gnn_attn_drop,
                                        emb_rnn_dropout=emb_rnn_dropout, emb_fix_word_emb=emb_fix_word_emb,
                                        emb_fix_bert_emb=emb_fix_bert_emb,
                                        emb_word_dropout=emb_word_dropout,
                                        gnn_hidden_size=gnn_hidden_size, gnn_input_size=gnn_input_size,
                                        gnn_output_size=gnn_output_size,
                                        **kwargs)

        self.use_copy = dec_use_copy
        self.use_coverage = dec_use_coverage
        self.dec_rnn_type = dec_rnn_type
        self.share_vocab = share_vocab

        self._build_decoder(rnn_type=dec_rnn_type, decoder_length=dec_max_decoder_step, vocab_model=vocab_model,
                            rnn_input_size=emb_hidden_size, share_vocab=share_vocab,
                            input_size=2 * gnn_hidden_size if gnn_direction_option == 'bi_sep' else gnn_hidden_size,
                            hidden_size=dec_hidden_size, graph_pooling_strategy=dec_graph_pooling_strategy,
                            use_copy=dec_use_copy, use_coverage=dec_use_coverage,
                            tgt_emb_as_output_layer=dec_tgt_emb_as_output_layer,
                            attention_type=dec_attention_type, node_type_num=dec_node_type_num,
                            fuse_strategy=dec_fuse_strategy, teacher_forcing_rate=dec_teacher_forcing_rate,
                            fix_word_emb=emb_fix_word_emb,
                            rnn_dropout=dec_dropout)

    def _build_decoder(self, decoder_length, input_size, rnn_input_size, hidden_size, graph_pooling_strategy,
                       vocab_model, fix_word_emb=False, share_vocab=False,
                       use_copy=False, use_coverage=False, tgt_emb_as_output_layer=False, teacher_forcing_rate=1.0,
                       rnn_type="lstm", attention_type="uniform", node_type_num=None, fuse_strategy="average",
                       rnn_dropout=0.2):
        if share_vocab and self.enc_word_emb is not None:
            self.dec_word_emb = self.enc_word_emb.word_emb_layer
        else:
            self.dec_word_emb = WordEmbedding(vocab_model.out_word_vocab.embeddings.shape[0],
                                              vocab_model.out_word_vocab.embeddings.shape[1],
                                              pretrained_word_emb=vocab_model.out_word_vocab.embeddings,
                                              fix_emb=fix_word_emb).word_emb_layer

        self.seq_decoder = StdRNNDecoder(rnn_type=rnn_type, max_decoder_step=decoder_length,
                                         input_size=input_size,
                                         hidden_size=hidden_size, graph_pooling_strategy=graph_pooling_strategy,
                                         word_emb=self.dec_word_emb, vocab=vocab_model.out_word_vocab,
                                         attention_type=attention_type, fuse_strategy=fuse_strategy,
                                         node_type_num=node_type_num,
                                         rnn_emb_input_size=rnn_input_size, use_coverage=use_coverage,
                                         use_copy=use_copy,
                                         tgt_emb_as_output_layer=tgt_emb_as_output_layer, dropout=rnn_dropout)
        self.teacher_forcing_rate = teacher_forcing_rate

    def encoder_decoder(self, batch_graph, oov_dict=None, tgt_seq=None):
        # run GNN
        batch_graph = self.gnn_encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']

        # down-task
        prob, enc_attn_weights, coverage_vectors = self.seq_decoder(batch_graph, tgt_seq=tgt_seq, teacher_forcing_rate=self.teacher_forcing_rate,
                                                                    oov_dict=oov_dict)
        return prob, enc_attn_weights, coverage_vectors

    def encoder_decoder_beam_search(self, batch_graph, beam_size, topk=1, oov_dict=None):
        generator = DecoderStrategy(beam_size=beam_size, vocab=self.seq_decoder.vocab, rnn_type=self.dec_rnn_type,
                                    decoder=self.seq_decoder, use_copy=self.use_copy,
                                    use_coverage=self.use_coverage, max_decoder_step=self.seq_decoder.max_decoder_step)

        batch_graph = self.gnn_encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']
        beam_results = generator.generate(batch_graph=batch_graph, oov_dict=oov_dict, topk=topk)
        return beam_results

    def forward(self, batch_graph, tgt_seq=None, oov_dict=None):
        """
            The forward function of Graph2Seq model.
        Parameters
        ----------
        batch_graph: GraphData
            The graph input
        tgt_seq: torch.Tensor
            The target sequence
        oov_dict: VocabModel, default=None
            The vocabulary for copy mechanism.

        Returns
        -------
        logits: torch.Tensor
            shape=[B, tgt_len, vocab_size]
            The probability for predicted target sequence. It is processed by softmax function.
        enc_attn_weights_average: torch.Tensor
            It is used for calculating coverage loss.
            The averaged attention scores.
        coverage_vectors: torch.Tensor
            It is used for calculating coverage loss.
            The coverage vector.
        """
        batch_graph = self.graph_topology(batch_graph)
        return self.encoder_decoder(batch_graph=batch_graph, oov_dict=oov_dict, tgt_seq=tgt_seq)

    def translate(self, batch_graph, beam_size, topk=1, oov_dict=None):
        """
            Decoding with the support of beam_search.
            Specifically, when ``beam_size`` is 1, it is equal to greedy search.
        Parameters
        ----------
        batch_graph: GraphData
            The graph input
        beam_size: int
            The beam width. When it is 1, the output is equal to greedy search's output.
        topk: int, default=1
            The number of decoded sequence to be reserved.
            Usually, ``topk`` should be smaller or equal to ``beam_size``
        oov_dict: VocabModel, default=None
            The vocabulary for copy.

        Returns
        -------
        results: torch.Tensor
            The results with the shape of ``[batch_size, topk, max_decoder_step]`` containing the word indexes.
        """

        batch_graph = self.graph_topology(batch_graph)
        return self.encoder_decoder_beam_search(batch_graph=batch_graph, beam_size=beam_size, topk=topk, oov_dict=oov_dict)

    @classmethod
    def from_args(cls, opt, vocab_model):
        """
            The function for building ``Graph2Seq`` model.
        Parameters
        ----------
        opt: dict
            The configuration dict. It should has the same hierarchy and keys as the template.
        vocab_model: VocabModel
            The vocabulary.

        Returns
        -------
        model: Graph2Seq
        """
        emb_args = cls._get_node_embedding_params(opt)
        gnn_args = cls._get_gnn_params(opt)
        dec_args = cls._get_decoder_params(opt)

        args = (copy.deepcopy(emb_args))
        args.update(gnn_args)
        args.update(dec_args)
        args["share_vocab"] = opt["graph_construction_args"]["graph_construction_share"]["share_vocab"]

        return cls(vocab_model=vocab_model, **args)

    @staticmethod
    def _get_decoder_params(opt):
        dec_args = opt["decoder_args"]
        shared_args = copy.deepcopy(dec_args["rnn_decoder_share"])
        private_args = copy.deepcopy(dec_args["rnn_decoder_private"])
        ret = copy.deepcopy(dict(shared_args, **private_args))
        dec_ret = {"dec_" + key: value for key, value in ret.items()}
        return dec_ret

    @staticmethod
    def _get_gnn_params(opt):
        args = opt["graph_embedding_args"]
        shared_args = copy.deepcopy(args["graph_embedding_share"])
        private_args = copy.deepcopy(args["graph_embedding_private"])
        if "activation" in private_args.keys():
            private_args["activation"] = getattr(F, private_args["activation"]) if private_args["activation"] else None
        if "norm" in private_args.keys():
            private_args["norm"] = getattr(F, private_args["norm"]) if private_args["norm"] else None

        gnn_shared_args = {"gnn_" + key: value for key, value in shared_args.items()}
        pri_shared_args = {"gnn_" + key: value for key, value in private_args.items()}
        ret = copy.deepcopy(dict(gnn_shared_args, **pri_shared_args))
        ret["gnn"] = opt["graph_embedding_name"]
        return ret

    @staticmethod
    def _get_node_embedding_params(opt):
        args = opt["graph_construction_args"]["node_embedding"]
        ret: dict = copy.deepcopy(args)
        ret.pop("embedding_style")
        emb_ret = {"emb_" + key: value for key, value in ret.items()}
        emb_ret["embedding_style"] = args["embedding_style"]
        emb_ret["graph_type"] = opt["graph_construction_name"]
        return emb_ret



import collections
from torch import nn
import torch as th
from ..base import NodeClassifierLayerBase

class FeedForwardNNLayer(NodeClassifierLayerBase):
    r"""Specific class for node classification task.


    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    num_class : int 
               The number of node catrgoriey for classification
    hidden_size : list of int type values
                  Example for two layers's FeedforwardNN: [50, 20]
    activation: the activation function class for each fully connected layer
                Default: nn.ReLU()
                Example: nn.ReLU(),nn.Sigmoid().        

    """     
    def __init__(self, input_size, num_class, hidden_size,activation=nn.ReLU()):        
        super(FeedForwardNNLayer, self).__init__()
           
            
        #build the linear module list
        module_seq=[]        
        
        for layer_idx in range(len(hidden_size)):
            if layer_idx==0:
                module_seq.append(('linear'+str(layer_idx),nn.Linear(input_size,hidden_size[layer_idx])))
            else:
                module_seq.append(('linear'+str(layer_idx),nn.Linear(hidden_size[layer_idx-1],self.hidden_size[layer_idx])))
            module_seq.append(('activate'+str(layer_idx),activation))
            
        module_seq.append(('linear_end',nn.Linear(hidden_size[-1],num_class)))
        
        self.classifier = nn.Sequential(collections.OrderedDict(module_seq))

    def forward(self, node_emb, node_idx=None):
        r"""
        Forward functions to compute the logits tensor for node classification.
    
      
    
        Parameters
        ----------
    
        node_emb : tensor [N,H]  
                   N: number of nodes    
                   H: length of the node embeddings
        node_idx : a list of index of nodes that needs classification.
                   Default: 'None'
    
        Returns 
        -------
             logit tensor: [N, num_class] The score logits for all nodes preidcted.
        """ 
        if node_idx == None:
            return self.classifier(node_emb)
        else:
            new_emb_new = node_emb[th.tensor(node_idx), :]  # get the required node embeddings.
            return self.classifier(new_emb_new)



import torch
import torch.nn as nn
from .base import KGLossBase


class SigmoidLoss(nn.Module):
    r"""
    Pairwise loss function. :math:`x_p` is the predition score of positive examples,
    and :math:`x_n` is the predition score of negative examples.

    .. math::
        \text{loss}(x_p, x_n) = -{\sum_i log \frac{1} {1+\exp(-x_p[i])} +
        \sum_j \frac{1} log(1+\exp(x_n[j])))
    """
    def __init__(self, adv_temperature = None):
        super(SigmoidLoss, self).__init__()
        self.criterion = nn.LogSigmoid()
        if adv_temperature != None:
            self.adv_temperature = nn.Parameter(torch.Tensor([adv_temperature]))
            self.adv_temperature.requires_grad = False
            self.adv_flag = True
        else:
            self.adv_flag = False

    def get_weights(self, n_score):
        return torch.softmax(n_score * self.adv_temperature, dim = -1).detach()

    def forward(self, p_score, n_score):
        if self.adv_flag:
            return -(self.criterion(p_score).mean() + (self.get_weights(n_score) * self.criterion(-n_score)).sum(dim = -1).mean()) / 2
        else:
            return -(self.criterion(p_score).mean() + self.criterion(-n_score).mean()) / 2


class SoftplusLoss(nn.Module):
    r"""
    Pairwise loss function. :math:`x_p` is the predition score of positive examples,
    and :math:`x_n` is the predition score of negative examples.

    .. math::
        \text{loss}(x_p, x_n) = \sum_i log(1+\exp(-x_p[i])) + \sum_j log(1+\exp(x_n[j]))
    """
    def __init__(self, adv_temperature=None):
        super(SoftplusLoss, self).__init__()
        self.criterion = nn.Softplus()
        if adv_temperature != None:
            self.adv_temperature = nn.Parameter(torch.Tensor([adv_temperature]))
            self.adv_temperature.requires_grad = False
            self.adv_flag = True
        else:
            self.adv_flag = False

    def get_weights(self, n_score):
        return torch.softmax(n_score * self.adv_temperature, dim=-1).detach()

    def forward(self, p_score, n_score):
        if self.adv_flag:
            return (self.criterion(-p_score).mean() + (self.get_weights(n_score) * self.criterion(n_score)).sum(
                dim=-1).mean()) / 2
        else:
            return (self.criterion(-p_score).mean() + self.criterion(n_score).mean()) / 2


class KGLoss(KGLossBase):
    r"""
    In the state-of-the-art KGE models, loss functions were designed according to
    various pointwise, pairwise and multi-class approaches. Refers to
    `Loss Functions in Knowledge Graph Embedding Models <https://alammehwish.github.io/dl4kg-eswc/papers
    /paper%201.pdf>`__

    **Pointwise Loss Function**

    `MSELoss <https://pytorch.org/docs/master/generated/torch.nn.MSELoss.html>`__
    Creates a criterion that measures the mean squared error (squared L2 norm)
    between each element in the input :math:`x` and target :math:`y`.


    `SOFTMARGINLOSS <https://pytorch.org/docs/master/generated/torch.nn.SoftMarginLoss
    .html>`__ Creates a criterion that optimizes a two-class classification
    logistic loss between input tensor :math:`x` and target tensor :math:`y`
    (containing 1 or -1).
    Tips:   The number of positive and negative samples should be about the same,
            otherwise it's easy to overfit

    .. math::
        \text{loss}(x, y) = \sum_i \frac{\log(1 + \exp(-y[i]*x[i]))}{\text{x.nelement}()}


    **Pairwise Loss Function**

    `SoftplusLoss <https://github.com/thunlp/OpenKE/blob/OpenKE-PyTorch/openke/module/loss/SoftplusLoss.py>`__
    refers to the paper `OpenKE: An Open Toolkit for Knowledge Embedding <https://www.aclweb.org/anthology/D18-2024.pdf>`__

    `SigmoidLoss <https://github.com/thunlp/OpenKE/blob/OpenKE-PyTorch/openke/module/loss/SigmoidLoss.py>`__
    refers to the paper `OpenKE: An Open Toolkit for Knowledge Embedding <https://www.aclweb.org/anthology/D18-2024.pdf>`__

    **Multi-Class Loss Function**

    `Binary Cross Entropy Loss <https://pytorch.org/docs/master/generated/torch.nn.BCELoss.html>`__
    Creates a criterion that measures the Binary Cross Entropy between the target and the output. Note that the targets
    :math:`y` should be numbers between 0 and 1.

    """
    def __init__(self,
                 loss_type,
                 size_average=None,
                 reduce=None,
                 reduction='mean',
                 adv_temperature=None,
                 weight=None):
        super(KGLoss, self).__init__()
        self.loss_type = loss_type
        if loss_type == 'MSELoss':
            self.loss_function = nn.MSELoss(size_average, reduce, reduction)
        elif loss_type == 'SoftMarginLoss':
            self.loss_function = nn.SoftMarginLoss(size_average, reduce, reduction)
        elif loss_type == 'SoftplusLoss':
            self.loss_function = SoftplusLoss(adv_temperature)
        elif loss_type == 'SigmoidLoss':
            self.loss_function = SigmoidLoss(adv_temperature)
        elif loss_type == 'BCELoss':
            self.loss_function = nn.BCELoss(weight, size_average, reduce, reduction)
        else:
            raise NotImplementedError()
        return

    def forward(self, input=None, target=None, p_score=None, n_score=None):
        """

        Parameters
        ----------
        MSELoss
            input: Tensor.
                :math:`(N,*)` where :math:`*` means any number of additional dimensions
            target: Tensor.
                :math:`(N,*)`, same shape as the input
            output:
                If reduction is `'none'`, then same shape as the input

        SoftMarginLoss
            input: Tensor.
                :math:`(*)` where * means, any number of additional dimensions
            target: Tensor.
                same shape as the input
            output: scalar.
                If reduction is `'none'`, then same shape as the input

        SoftplusLoss
            p_score: Tensor.
                :math:`(*)` where * means, any number of additional dimensions
            n_score: Tensor.
                :math:`(*)` where * means, any number of additional dimensions.
                The dimension could be different from the `p_score` dimension.
            output: scalar.

        SigmoidLoss
            p_score: Tensor.
                :math:`(*)` where * means, any number of additional dimensions
            n_score: Tensor.
                :math:`(*)` where * means, any number of additional dimensions.
                The dimension could be different from the `p_score` dimension.
            output: scalar.

        BCELoss:
            Input: Tensor.
                :math:`(N, *)` where :math:`*` means, any number of additional dimensions
            Target: Tensor.
                :math:`(N, *)`, same shape as the input
            Output: scalar.
                If :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as input.

        Returns
        -------

        """
        if self.loss_type in ['SoftplusLoss', 'SigmoidLoss']:
            return self.loss_function(p_score, n_score)
        else:
            return self.loss_function(input, target)


from ..base import KGCompletionBase
from .ComplExLayer import ComplExLayer
from .....data.data import GraphData


class ComplEx(KGCompletionBase):
    r"""Specific class for knowledge graph completion task.

    ComplEx from paper `Complex Embeddings for Simple Link Prediction
    <http://proceedings.mlr.press/v48/trouillon16.pdf>`__.


    Parameters
    ----------
    input_dropout: float
        Dropout for node_emb and rel_emb. Default: `0.0`

    loss_name: str
        The loss type selected fot the KG completion task. Default: `'BCELoss'`
    """

    def __init__(self,
                 input_dropout=0.0,
                 loss_name='BCELoss'):
        super(ComplEx, self).__init__()
        self.loss_name = loss_name
        self.classifier = ComplExLayer(input_dropout, loss_name)


    def forward(self, input_graph: GraphData,
                e1_embedded_real, rel_embedded_real,
                e1_embedded_img, rel_embedded_img,
                all_node_emb_real, all_node_emb_img,
                multi_label=None):
        r"""
        Forward functions to compute the logits tensor for kg completion.

        Parameters
        ----------

        input graph : GraphData
            The tensors stored in the node feature field named "node_emb" and
            "rel_emb" in the input_graph are used for knowledge graph completion.

        e1_embedded_real : tensor [B, H]
            The selected entity_1 real embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_embedded_real : tensor [B, H]
            The selected relation real embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        e1_embedded_img : tensor [B, H]
            The selected entity_1 img embeddings of a batch.
            B: batch size
            H: length of the node embeddings (entity embeddings)

        rel_embedded_img : tensor [B, H]
            The selected relation img embeddings of a batch.
            B: batch size
            H: length of the edge embeddings (relation embeddings)

        all_node_emb_real :  torch.nn.modules.sparse.Embedding [N, H]
            All node real embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node real embeddings (entity embeddings)

        all_node_emb_img :  torch.nn.modules.sparse.Embedding [N, H]
            All node img embeddings.
            N: number of nodes in the whole KG graph
            H: length of the node img embeddings (entity embeddings)

        multi_label: tensor [B, N]
            multi_label is a binary matrix. Each element can be equal to 1 for true label
            and 0 for false label (or 1 for true label, -1 for false label).
            multi_label[i] represents a multi-label of a given head-rel pair.
            B is the batch size.
            N: number of nodes in the whole KG graph.

        Returns
        ---------

        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class]
        """

        if multi_label is None:
            input_graph.graph_attributes['logits'] = self.classifier(e1_embedded_real,
                                                                     e1_embedded_img,
                                                                     rel_embedded_real,
                                                                     rel_embedded_img,
                                                                     all_node_emb_real,
                                                                     all_node_emb_img)
        else:
            input_graph.graph_attributes['logits'], input_graph.graph_attributes['p_score'], \
            input_graph.graph_attributes['n_score'] = self.classifier(e1_embedded_real,
                                                                      e1_embedded_img,
                                                                      rel_embedded_real,
                                                                      rel_embedded_img,
                                                                      all_node_emb_real,
                                                                      all_node_emb_img,
                                                                      multi_label)

        return input_graph


import yaml
from collections import OrderedDict
import numpy as np
from scipy import sparse
import torch
import torch.nn as nn



def get_config(config_path="config.yml"):
    with open(config_path, "r") as setting:
        config = yaml.load(setting)
    return config

def print_config(config):
    print("**************** MODEL CONFIGURATION ****************")
    for key in sorted(config.keys()):
        val = config[key]
        keystr = "{}".format(key) + (" " * (24 - len(key)))
        print("{} -->   {}".format(keystr, val))
    print("**************** MODEL CONFIGURATION ****************")

def grid(kwargs):
    """Builds a mesh grid with given keyword arguments for this Config class.
    If the value is not a list, then it is considered fixed"""

    class MncDc:
        """This is because np.meshgrid does not always work properly..."""

        def __init__(self, a):
            self.a = a  # tuple!

        def __call__(self):
            return self.a

    def merge_dicts(*dicts):
        """
        Merges dictionaries recursively. Accepts also `None` and returns always a (possibly empty) dictionary
        """
        from functools import reduce
        def merge_two_dicts(x, y):
            z = x.copy()  # start with x's keys and values
            z.update(y)  # modifies z with y's keys and values & returns None
            return z

        return reduce(lambda a, nd: merge_two_dicts(a, nd if nd else {}), dicts, {})


    sin = OrderedDict({k: v for k, v in kwargs.items() if isinstance(v, list)})
    for k, v in sin.items():
        copy_v = []
        for e in v:
            copy_v.append(MncDc(e) if isinstance(e, tuple) else e)
        sin[k] = copy_v

    grd = np.array(np.meshgrid(*sin.values()), dtype=object).T.reshape(-1, len(sin.values()))
    return [merge_dicts(
        {k: v for k, v in kwargs.items() if not isinstance(v, list)},
        {k: vv[i]() if isinstance(vv[i], MncDc) else vv[i] for i, k in enumerate(sin)}
    ) for vv in grd]


def to_cuda(x, device=None):
    if device:
        x = x.to(device)

    return x

def create_mask(x, N, device=None):
    if isinstance(x, torch.Tensor):
        x = x.data
    mask = np.zeros((len(x), N))
    for i in range(len(x)):
        mask[i, :x[i]] = 1

    return to_cuda(torch.Tensor(mask), device)

def normalize_adj(mx):
    """Row-normalize matrix: symmetric normalized Laplacian"""
    rowsum = mx.sum(1)
    r_inv_sqrt = torch.pow(rowsum, -0.5).flatten()
    r_inv_sqrt[torch.isinf(r_inv_sqrt)] = 0.
    r_mat_inv_sqrt = torch.diag(r_inv_sqrt)

    return torch.mm(torch.mm(mx, r_mat_inv_sqrt).transpose(-1, -2), r_mat_inv_sqrt)


# def normalize_adj(mx):
#     """Normalize matrix: asymmetric normalized Laplacian"""
#     rowsum = mx.sum(1)
#     r_inv_sqrt = torch.pow(rowsum, -0.5).flatten()
#     r_inv_sqrt[torch.isinf(r_inv_sqrt)] = 0.
#     r_mat_inv_sqrt = torch.diag(r_inv_sqrt)

#     colsum = mx.sum(0)
#     c_inv_sqrt = torch.pow(colsum, -0.5).flatten()
#     c_inv_sqrt[torch.isinf(c_inv_sqrt)] = 0.
#     c_mat_inv_sqrt = torch.diag(c_inv_sqrt)

#     return torch.mm(torch.mm(r_mat_inv_sqrt, mx), c_mat_inv_sqrt)

def normalize_sparse_adj(mx):
    """symmetric normalized Laplacian"""
    rowsum = np.array(mx.sum(1))
    r_inv_sqrt = np.power(rowsum, -0.5).flatten()
    r_inv_sqrt[np.isinf(r_inv_sqrt)] = 0.
    r_mat_inv_sqrt = sparse.diags(r_inv_sqrt)

    return mx.dot(r_mat_inv_sqrt).transpose().dot(r_mat_inv_sqrt)

def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)

    return torch.sparse.FloatTensor(indices, values, shape)

def dropout_fn(x, drop_prob, shared_axes=[], training=False):
    """
    Apply dropout to input tensor.
    Parameters
    ----------
    input_tensor: ``torch.FloatTensor``
        A tensor of shape ``(batch_size, ..., num_timesteps, embedding_dim)``
    Returns
    -------
    output: ``torch.FloatTensor``
        A tensor of shape ``(batch_size, ..., num_timesteps, embedding_dim)`` with dropout applied.
    """
    if drop_prob == 0 or drop_prob == None or (not training):
        return x

    sz = list(x.size())
    for i in shared_axes:
        sz[i] = 1
    mask = x.new(*sz).bernoulli_(1. - drop_prob).div_(1. - drop_prob)
    mask = mask.expand_as(x)

    return x * mask

class Identity(nn.Module):
    """A placeholder identity operator that is argument-insensitive.
    (Identity has already been supported by PyTorch 1.2, we will directly
    import torch.nn.Identity in the future)
    """
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        """Return input"""
        return x

class EarlyStopping:
    def __init__(self, save_model_path, patience=10):
        self.patience = patience
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.save_model_path = save_model_path

    def step(self, acc, model):
        score = acc
        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(model)
        elif score < self.best_score:
            self.counter += 1
            print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(model)
            self.counter = 0

        return self.early_stop

    def save_checkpoint(self, model):
        '''Saves model when validation loss decrease.'''
        torch.save(model.state_dict(), self.save_model_path)
        print('Saved model to {}'.format(self.save_model_path))

    def load_checkpoint(self, model):
        model.load_state_dict(torch.load(self.save_model_path))

        return model



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from graph4nlp.pytorch.modules.prediction.generation.TreeBasedDecoder import StdTreeDecoder
from graph4nlp.pytorch.modules.utils.tree_utils import Vocab

from graph4nlp.pytorch.modules.graph_construction import *
from graph4nlp.pytorch.modules.graph_embedding import *


class Graph2Tree(nn.Module):
    def __init__(self, src_vocab,
                 tgt_vocab,
                 use_copy,
                 gnn_type,
                 direction_option,
                 graph_construction_type,
                 enc_hidden_size,
                 dec_hidden_size,
                 dropout_for_word_embedding,
                 dropout_for_encoder_feature,
                 dropout_for_encoder_attn,
                 dropout_for_decoder,
                 device,
                 criterion,
                 teacher_force_ratio,
                 max_dec_seq_length,
                 max_dec_tree_depth,
                 embedding_style,
                 K,
                 gat_head,
                 sage_aggr,
                 attn_type,
                 use_sibling,
                 use_share_vocab):
        super(Graph2Tree, self).__init__()

        self.src_vocab = src_vocab
        self.tgt_vocab = tgt_vocab
        self.device = device
        self.use_copy = use_copy
        self.input_size = self.src_vocab.vocab_size
        self.output_size = self.tgt_vocab.vocab_size
        self.criterion = criterion
        self.use_share_vocab = use_share_vocab

        if graph_construction_type == "DependencyGraph":
            self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                   vocab=self.src_vocab,
                                                                   hidden_size=enc_hidden_size, 
                                                                   word_dropout=dropout_for_word_embedding, 
                                                                   rnn_dropout=dropout_for_word_embedding, 
                                                                   fix_word_emb=False)
            self.use_edge_weight = False
        elif graph_construction_type == "ConstituencyGraph":
            self.graph_topology = ConstituencyBasedGraphConstruction(embedding_style=embedding_style,
                                                                     vocab=self.src_vocab,
                                                                     hidden_size=enc_hidden_size,
                                                                     word_dropout=dropout_for_word_embedding,
                                                                     rnn_dropout=dropout_for_word_embedding,
                                                                     fix_word_emb=False)
            self.use_edge_weight = False
        elif graph_construction_type == "DynamicGraph_node_emb":
            self.graph_topology = NodeEmbeddingBasedGraphConstruction(self.src_vocab,
                                                                      embedding_style,
                                                                      sim_metric_type='weighted_cosine',
                                                                      num_heads=1,
                                                                      top_k_neigh=None,
                                                                      epsilon_neigh=0.5,
                                                                      smoothness_ratio=0.1,
                                                                      connectivity_ratio=0.05,
                                                                      sparsity_ratio=0.1,
                                                                      input_size=enc_hidden_size,
                                                                      hidden_size=enc_hidden_size,
                                                                      fix_word_emb=False,
                                                                      word_dropout=dropout_for_word_embedding,
                                                                      rnn_dropout=dropout_for_word_embedding
                                                                      )
            self.use_edge_weight = True
        elif graph_construction_type == "DynamicGraph_node_emb_refined":
            self.graph_topology = NodeEmbeddingBasedRefinedGraphConstruction(self.src_vocab,
                                                                             embedding_style,
                                                                             0.2,
                                                                             sim_metric_type="weighted_cosine",
                                                                             num_heads=1,
                                                                             top_k_neigh=None,
                                                                             epsilon_neigh=0.5,
                                                                             smoothness_ratio=0.1,
                                                                             connectivity_ratio=0.05,
                                                                             sparsity_ratio=0.1,
                                                                             input_size=enc_hidden_size,
                                                                             hidden_size=enc_hidden_size,
                                                                             fix_word_emb=False,
                                                                             word_dropout=dropout_for_word_embedding,
                                                                             rnn_dropout=dropout_for_word_embedding,
                                                                             device=device)
            self.use_edge_weight = True
        else:
            raise NotImplementedError()

        self.word_emb = self.graph_topology.embedding_layer.word_emb_layers['w2v'].word_emb_layer
        if self.use_share_vocab == 0:
            self.tgt_word_embedding = nn.Embedding(self.tgt_vocab.vocab_size, dec_hidden_size, 
                                                padding_idx=self.tgt_vocab.get_symbol_idx(self.tgt_vocab.pad_token),
                                                _weight=torch.from_numpy(self.tgt_vocab.embeddings).float())

        if gnn_type == "GAT":
            self.encoder = GAT(K, enc_hidden_size, enc_hidden_size, enc_hidden_size, gat_head,
                               direction_option=direction_option, feat_drop=dropout_for_encoder_feature,
                               attn_drop=dropout_for_encoder_attn, activation=F.relu, residual=True)
        elif gnn_type == "GGNN":
            self.encoder = GGNN(K, enc_hidden_size, enc_hidden_size, enc_hidden_size,
                                feat_drop=dropout_for_encoder_feature, use_edge_weight=self.use_edge_weight,
                                direction_option=direction_option)
        elif gnn_type == "SAGE":
            self.encoder = GraphSAGE(K, enc_hidden_size, enc_hidden_size, enc_hidden_size,
                                     sage_aggr, direction_option=direction_option, feat_drop=dropout_for_encoder_feature,
                                     activation=F.relu, bias=True, use_edge_weight=self.use_edge_weight)
        elif gnn_type == "GCN":
            self.encoder = GCN(K, enc_hidden_size, enc_hidden_size, enc_hidden_size,
                               direction_option=direction_option, gcn_norm="both", activation=F.relu,
                               use_edge_weight=self.use_edge_weight)
        else:
            print("Wrong gnn type, please use GCN GAT GGNN or SAGE")
            raise NotImplementedError()

        self.decoder = StdTreeDecoder(attn_type=attn_type,
                                      embeddings=self.word_emb if self.use_share_vocab else self.tgt_word_embedding,
                                      enc_hidden_size=enc_hidden_size,
                                      dec_emb_size=self.tgt_vocab.embedding_dims,
                                      dec_hidden_size=dec_hidden_size,
                                      output_size=self.output_size,
                                      device=device,
                                      criterion=self.criterion,
                                      teacher_force_ratio=teacher_force_ratio,
                                      use_sibling=use_sibling,
                                      use_copy=self.use_copy,
                                      dropout_for_decoder=dropout_for_decoder,
                                      max_dec_seq_length=max_dec_seq_length,
                                      max_dec_tree_depth=max_dec_tree_depth,
                                      tgt_vocab=self.tgt_vocab)

    def forward(self, batch_graph, tgt_tree_batch, oov_dict=None):
        batch_graph = self.graph_topology(batch_graph)
        batch_graph = self.encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']

        loss = self.decoder(g=batch_graph, tgt_tree_batch=tgt_tree_batch, oov_dict=oov_dict)
        return loss

    def init(self, init_weight):
        self.encoder.to(self.device)
        self.decoder.to(self.device)

        for name, param in self.named_parameters():
            if param.requires_grad:
                if ("word_embedding" in name) or ("word_emb_layer" in name) or ("bert_embedding" in name):
                    pass
                else:
                    if len(param.size()) >= 2:
                        if "rnn" in name:
                            init.orthogonal_(param)
                        else:
                            init.xavier_uniform_(param, gain=1.0)
                    else:
                        init.uniform_(param, -init_weight, init_weight)
    @classmethod
    def from_args(cls, opt, src_vocab, tgt_vocab, device, embedding_style, criterion):
        return cls(src_vocab=src_vocab, tgt_vocab=tgt_vocab, use_copy=opt.use_copy, gnn_type=opt.gnn_type, 
                    direction_option=opt.direction_option, graph_construction_type=opt.graph_construction_type,
                    enc_hidden_size=opt.enc_hidden_size, dec_hidden_size=opt.dec_hidden_size, 
                    dropout_for_word_embedding=opt.dropout_for_word_embedding, dropout_for_encoder_feature=opt.dropout_for_encoder,
                    dropout_for_encoder_attn=opt.dropout_for_encoder, dropout_for_decoder=opt.dropout_for_decoder, 
                    device=device, criterion=criterion, teacher_force_ratio=opt.teacher_force_ratio, max_dec_seq_length=opt.max_dec_seq_length, 
                    max_dec_tree_depth=opt.max_dec_tree_depth, embedding_style=embedding_style, K=opt.K, 
                    gat_head=[int(i) for i in opt.gat_head.split(',')], sage_aggr=opt.sage_aggr, 
                    attn_type=opt.attn_type, use_sibling=opt.use_sibling, use_share_vocab=opt.use_share_vocab)


from ....modules.utils.config_utils import get_yaml_config
import os

str2yaml = {"gat": "gat.yaml", "gcn": "gcn.yaml", "ggnn": "ggnn.yaml", "graphsage": "graphsage.yaml"}

dir_path = os.path.dirname(os.path.realpath(__file__))

def get_graph_embedding_args(graph_embedding_name):
    """
        It will build the template for ``GNNBase`` model.
    Parameters
    ----------
    graph_embedding_name: str
        The graph embedding name. Expected in ["gcn", "gat", "graphsage", "ggnn"].
        If it can't find the ``graph_embedding_name``, it will return ``{}``.
    Returns
    -------
    template_dict: dict
        The template dict.
        The structure is shown as follows:
        {
            graph_embedding_share: {num_layers: 1, input_size: 300, ...},
            graph_embedding_private: {heads: [1], attn_drop: 0.0}
        }
        The ``graph_embedding_share`` contains the parameters shared by all ``GNNBase`` models.
        The ``graph_embedding_private`` contains the parameters specifically in each graph_embedding methods.
    """
    if graph_embedding_name in str2yaml.keys():
        yaml_name = str2yaml[graph_embedding_name]
        path = os.path.join(dir_path, yaml_name)
        config = get_yaml_config(path)
        return config
    else:
        return {}


__all__ = ["get_graph_embedding_args"]



import os
import json
import sys
from . import constants as Constants


class Logger:
    def __init__(self, dirname, config=None, overwrite=False, logging=True):
        self.logging = logging
        if os.path.exists(dirname):
            if not overwrite:
                raise Exception('Directory already exists: {}'.format(dirname))
        else:
            os.makedirs(dirname)

        if config is not None:
            self.log_json(config, os.path.join(dirname, Constants._CONFIG_FILE))

        if logging:
            self.fout = open(os.path.join(dirname, Constants._SAVED_METRICS_FILE), 'a')

    def log_json(self, data, filename, mode='w'):
        with open(filename, mode) as outfile:
            outfile.write(json.dumps(data, indent=4, ensure_ascii=False))

    def write(self, text):
        if self.logging:
            self.fout.writelines(text + '\n')
            self.fout.flush()

    def close(self):
        if self.logging:
            self.fout.close()



import torch
import torch.nn as nn
import torch.nn.functional as F

import random
import numpy as np

from .base import RNNTreeDecoderBase
from ...utils.tree_utils import Tree, to_cuda
from .attention import Attention

from graph4nlp.pytorch.data.data import GraphData, from_batch
from graph4nlp.pytorch.modules.utils.tree_utils import to_cuda
from graph4nlp.pytorch.modules.prediction.generation.decoder_strategy import DecoderStrategy


class StdTreeDecoder(RNNTreeDecoderBase):
    r"""StdTreeDecoder: This is a tree decoder implementation, which is used for tree object decoding.

    Attributes
    ----------
    attn_type : str,
        Describe which attention mechanism is used, can be ``uniform``, ``separate_on_encoder_type``, ``separate_on_node_type``.

    embeddings : torch.nn.Module,
        Embedding layer, input is tensor of word index, output is word embedding tensor.

    enc_hidden_size : int, 
        Size of encoder hidden state.

    dec_emb_size : int,
        Size of decoder word embedding layer output size.

    dec_hidden_size : int,
        Size of decoder hidden state. (namely the ``lstm`` or ``gru`` hidden size when rnn unit has been specified)

    output_size : int,
        Size of output vocabulary size.

    device : object,
        Device where parameters and data are store, a torch.device object.

    teacher_force_ratio : float,
        The ratio of possibility to use teacher force training.

    use_sibling : boolean,
        Whether feed sibling state in each decoding step.

    use_copy : boolean,
        Whether use copy mechanism in decoding.

    fuse_strategy: str, option=[None, "average", "concatenate"], default=None
        The strategy to fuse attention results generated by separate attention.
        "None": If we do ``uniform`` attention, we will set it to None.
        "``average``": We will take an average on all results.
        "``concatenate``": We will concatenate all results to one.

    num_layers : int, optional,
        Layer number of decoder rnn unit.

    dropout_for_decoder: float,
        Dropout ratio for decoder(include both the dropout for word embedding and the dropout for attention layer)
    
    tgt_vocab : object,
        The vocab object used in decoder, including all the word<->id pairs appeared in the output sentences.
    
    graph_pooling_strategy : str,
        The graph pooling strategy used to generate the graph embedding with node embeddings

    rnn_type: str, optional,
        The rnn unit is used, option=["lstm", "gru"], default="lstm".

    max_dec_seq_length : int, optional,
        In decoding, the decoding steps upper limit.

    max_dec_tree_depth : int, optional,
        In decoding, the tree depth lower limit.
    """

    def __init__(self, attn_type, embeddings, enc_hidden_size, dec_emb_size,
                 dec_hidden_size, output_size, device, criterion, teacher_force_ratio,
                 use_sibling=True, use_attention=True, use_copy=False,
                 fuse_strategy="average", num_layers=1,
                 dropout_for_decoder=0.1, rnn_type="lstm", max_dec_seq_length=512,
                 max_dec_tree_depth=256, tgt_vocab=None, graph_pooling_strategy="max"):

        super(StdTreeDecoder, self).__init__(use_attention=True,
                                             use_copy=use_copy,
                                             use_coverage=False,
                                             attention_type="uniform",
                                             fuse_strategy="average")
        self.num_layers = num_layers
        self.device = device
        self.criterion = criterion
        self.rnn_size = dec_hidden_size
        self.enc_hidden_size = enc_hidden_size
        self.hidden_size = dec_hidden_size
        self.max_dec_seq_length = max_dec_seq_length
        self.max_dec_tree_depth = max_dec_tree_depth
        self.tgt_vocab = tgt_vocab
        self.teacher_force_ratio = teacher_force_ratio
        self.use_sibling = use_sibling
        self.dec_emb_size = dec_emb_size
        self.dropout_input = dropout_for_decoder
        self.embeddings = embeddings
        self.graph_pooling_strategy = graph_pooling_strategy

        self.attn_state = {}
        self.use_copy = use_copy
        self.attention = Attention(query_size=dec_hidden_size,
                                   memory_size=enc_hidden_size*2 if (enc_hidden_size*2 == dec_hidden_size) else enc_hidden_size,
                                   hidden_size=dec_hidden_size,
                                   has_bias=True, dropout=dropout_for_decoder,
                                   attention_funtion="dot")
        self.separate_attn = (attn_type != "uniform")

        if self.separate_attn:
            self.linear_att = nn.Linear(3*dec_hidden_size, dec_hidden_size)
        else:
            self.linear_att = nn.Linear(2*dec_hidden_size, dec_hidden_size)
        self.linear_out = nn.Linear(dec_hidden_size, output_size)
        self.dropout_attn = nn.Dropout(dropout_for_decoder)
        self.logsoftmax = nn.LogSoftmax(dim=1)

        if self.use_copy:
            ptr_size = self.embeddings.embedding_dim
            ptr_size += 4*self.rnn_size
            self.ptr = nn.Linear(ptr_size, 1)

        self.rnn = self._build_rnn(rnn_type=rnn_type, input_size=output_size, emb_size=dec_emb_size,
                                   hidden_size=dec_hidden_size, dropout_input=dropout_for_decoder, use_sibling=use_sibling, device=device)

    def _run_forward_pass(self, 
                          graph_node_embedding, 
                          graph_node_mask, 
                          rnn_node_embedding, 
                          graph_level_embedding,
                          graph_edge_embedding=None, 
                          graph_edge_mask=None, 
                          tgt_tree_batch=None, 
                          enc_batch=None, 
                          oov_dict=None):
        r"""
            The private calculation method for decoder.

        Parameters
        ----------
        enc_batch : torch.Tensor,
            The input batch : (Batch_size * Source sentence word index tensor).

        tgt_tree_batch:
            The target tree to generate : consists of (Batch_size * Tree object), each node in a Tree object is either a word index or a children Tree object.

        graph_node_embedding: torch.Tensor,
            The graph node embedding matrix of shape :math:`(B, N, D_{in})`

        graph_node_mask: torch.Tensor,
            The graph node type mask matrix of shape :math`(B, N)`

        rnn_node_embedding: torch.Tensor,
            The rnn encoded embedding matrix of shape :math`(B, N, D_{in})`

        graph_level_embedding: torch.Tensor,
            graph level embedding of shape :math`(B, D_{in})`

        graph_edge_embedding: torch.Tensor,
            graph edge embedding of shape :math`(B, N, D_{in})`

        graph_edge_mask: torch.Tensor,
            graph edge type embedding

        oov_dict: dict,
            vocab dict used in copy mechanism to incorporate some new words which have never appeared in vocab for input sentences in training set.
        """
        tgt_batch_size = len(tgt_tree_batch)

        enc_outputs = graph_node_embedding

        if graph_level_embedding == None:
            if self.graph_pooling_strategy == "max":
                graph_level_embedding = torch.max(graph_node_embedding, 1)[0]
            elif self.graph_pooling_strategy == "min":
                graph_level_embedding = torch.min(graph_node_embedding, 1)[0]
            elif self.graph_pooling_strategy == "mean":
                graph_level_embedding = torch.mean(graph_node_embedding, 1)
            else:
                raise NotImplementedError()

            graph_cell_state = graph_level_embedding
            graph_hidden_state = graph_level_embedding
        else:
            graph_cell_state, graph_hidden_state = graph_level_embedding

        rnn_node_embedding = torch.zeros_like(
            graph_node_embedding, requires_grad=False)
        rnn_node_embedding = to_cuda(rnn_node_embedding, self.device)

        cur_index = 1
        loss = 0

        dec_batch, queue_tree, max_index = get_dec_batch(
            tgt_tree_batch, tgt_batch_size, self.device, self.tgt_vocab)

        dec_state = {}
        for i in range(self.max_dec_tree_depth + 1):
            dec_state[i] = {}
            for j in range(self.max_dec_seq_length + 1):
                dec_state[i][j] = {}

        while (cur_index <= max_index):
            if cur_index > self.max_dec_tree_depth:
                break
            for j in range(1, 3):
                dec_state[cur_index][0][j] = torch.zeros(
                    (tgt_batch_size, self.rnn_size), dtype=torch.float, requires_grad=False, device=self.device)
                dec_state[cur_index][0][j] = to_cuda(
                    dec_state[cur_index][0][j], self.device)

            sibling_state = torch.zeros(
                (tgt_batch_size, self.rnn_size), dtype=torch.float, requires_grad=False)
            sibling_state = to_cuda(sibling_state, self.device)

            # with torch.no_grad():
            if cur_index == 1:
                for i in range(tgt_batch_size):
                    dec_state[1][0][1][i, :] = graph_cell_state[i]
                    dec_state[1][0][2][i, :] = graph_hidden_state[i]

            else:
                for i in range(1, tgt_batch_size+1):
                    if (cur_index <= len(queue_tree[i])):
                        par_index = queue_tree[i][cur_index - 1]["parent"]
                        child_index = queue_tree[i][cur_index -
                                                    1]["child_index"]

                        dec_state[cur_index][0][1][i-1,
                                                   :] = dec_state[par_index][child_index][1][i-1, :]
                        dec_state[cur_index][0][2][i-1,
                                                   :] = dec_state[par_index][child_index][2][i-1, :]

                    flag_sibling = False
                    for q_index in range(len(queue_tree[i])):
                        if (cur_index <= len(queue_tree[i])) and (q_index < cur_index - 1) and (queue_tree[i][q_index]["parent"] == queue_tree[i][cur_index - 1]["parent"]) and (queue_tree[i][q_index]["child_index"] < queue_tree[i][cur_index - 1]["child_index"]):
                            flag_sibling = True
                            sibling_index = q_index
                    if flag_sibling:
                        sibling_state[i - 1, :] = dec_state[sibling_index][dec_batch[sibling_index].size(
                            1) - 1][2][i - 1, :]

            parent_h = dec_state[cur_index][0][2]
            for i in range(dec_batch[cur_index].size(1) - 1):
                teacher_force = random.random() < self.teacher_force_ratio
                if teacher_force != True and i > 0:
                    input_word = pred.argmax(1)
                else:
                    input_word = dec_batch[cur_index][:, i]
                pred, rnn_state_iter, attn_scores = self.decode_step(tgt_batch_size=tgt_batch_size,
                                                                     dec_single_input=input_word,
                                                                     dec_single_state=(
                                                                         dec_state[cur_index][i][1], dec_state[cur_index][i][2]),
                                                                     memory=enc_outputs,
                                                                     parent_state=parent_h,
                                                                     oov_dict=oov_dict,
                                                                     enc_batch=enc_batch)

                dec_state[cur_index][i +
                                     1][1], dec_state[cur_index][i+1][2] = rnn_state_iter

                pred = torch.log(pred + 1e-31)
                loss += self.criterion(pred, dec_batch[cur_index][:, i+1])
            cur_index = cur_index + 1
        loss = loss / tgt_batch_size
        return loss

    def _filter_oov(self, tokens, vocab):
        r"""The function used to mask some oov word in word embedding layer."""
        ret = tokens.clone()
        ret[tokens >= vocab.vocab_size] = vocab.get_symbol_idx(vocab.unk_token)
        return ret

    def decode_step(self, tgt_batch_size, 
                          dec_single_input, 
                          dec_single_state, 
                          memory, 
                          parent_state, 
                          input_mask=None, 
                          memory_mask=None, 
                          memory_candidate=None, 
                          sibling_state=None, 
                          oov_dict=None, 
                          enc_batch=None):
        """The decoding function in tree decoder.

        Parameters
        ----------
        tgt_batch_size : int,
            batch size.
        dec_single_input : torch.Tensor,
            word id matrix for decoder input: [B, N].
        dec_single_state : torch.Tensor
            the rnn decoding hidden state: [B, N, D].
        memory : torch.Tensor
            the encoder output node embedding.
        parent_state : torch.Tensor
            the parent embedding used in parent feeding mechanism.
        input_mask : torch.Tensor, optional
            input mask, by default None
        memory_mask : torch.Tensor, optional
            mask for encoder output, by default None
        memory_candidate : torch.Tensor, optional
            encoder output used for separate attention mechanism, by default None
        sibling_state : torch.Tensor, optional
            sibling state for sibling feeding mechanism, by default None
        oov_dict : object, optional
            out-of-vocabulary object for copy mechanism, by default None
        enc_batch : torch.Tensor,
            The input batch : (Batch_size * Source sentence word index tensor).
        """
        dec_single_input = self._filter_oov(dec_single_input, self.tgt_vocab)
        rnn_state_c, rnn_state_h, dec_emb = self.rnn(
            dec_single_input, dec_single_state[0], dec_single_state[1], parent_state, sibling_state)
        
        attn_collect = []
        score_collect = []

        if self.separate_attn:
            pass
        else:
            context_vector, attn_scores = self.attention(
                query=rnn_state_h, memory=memory)
            attn_collect.append(context_vector)
            score_collect.append(attn_scores)

        pred = F.tanh(self.linear_att(
            torch.cat((context_vector, rnn_state_h), 1)))
        decoder_output = self.linear_out(self.dropout_attn(pred))
        if self.use_copy:
            assert enc_batch is not None
            assert oov_dict is not None
            output = torch.zeros(tgt_batch_size, oov_dict.vocab_size).to(self.device)
            attn_ptr = torch.cat(attn_collect, dim=-1)

            pgen_collect = [dec_emb, torch.cat((rnn_state_c, rnn_state_h), -1), attn_ptr]
            prob_ptr = torch.sigmoid(self.ptr(torch.cat(pgen_collect, -1)))
            prob_gen = 1 - prob_ptr
            gen_output = torch.softmax(decoder_output, dim=-1)

            ret = prob_gen * gen_output
            need_pad_length = len(oov_dict) - len(self.tgt_vocab)
            output = torch.cat((ret, ret.new_zeros((tgt_batch_size, need_pad_length))), dim=1)
            # output[:, :self.tgt_vocab.vocab_size] = ret

            ptr_output = attn_scores
            output.scatter_add_(1, enc_batch, prob_ptr * ptr_output)
            decoder_output = output
            # decoder_output = -F.threshold(-output, -1.0, -1.0)
        else:
            decoder_output = torch.softmax(decoder_output, dim=-1)

        return decoder_output, (rnn_state_c, rnn_state_h), attn_scores

    def translate(self, use_copy,
                  enc_hidden_size,
                  dec_hidden_size,
                  model,
                  input_graph_list,
                  word_manager,
                  form_manager,
                  device,
                  max_dec_seq_length,
                  max_dec_tree_depth,
                  use_beam_search=True,
                  beam_size=4,
                  oov_dict=None):
        input_graph_list.batch_node_features["token_id"] = input_graph_list.batch_node_features["token_id"].to(device)
        prev_c = torch.zeros((1, dec_hidden_size), requires_grad=False)
        prev_h = torch.zeros((1, dec_hidden_size), requires_grad=False)

        batch_graph = model.graph_topology(input_graph_list)
        batch_graph = model.encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']

        params = model.decoder._extract_params(batch_graph)
        graph_node_embedding = params['graph_node_embedding']
        if model.decoder.graph_pooling_strategy == "max":
            graph_level_embedding = torch.max(graph_node_embedding, 1)[0]
        rnn_node_embedding = params['rnn_node_embedding']
        graph_node_mask = params['graph_node_mask']
        enc_w_list = params['enc_batch']

        enc_outputs = graph_node_embedding
        prev_c = graph_level_embedding
        prev_h = graph_level_embedding

        # decode
        queue_decode = []
        queue_decode.append(
            {"s": (prev_c, prev_h), "parent": 0, "child_index": 1, "t": Tree()})
        head = 1
        while head <= len(queue_decode) and head <= max_dec_tree_depth:
            s = queue_decode[head-1]["s"]
            parent_h = s[1]
            t = queue_decode[head-1]["t"]

            sibling_state = torch.zeros(
                (1, dec_hidden_size), dtype=torch.float, requires_grad=False)
            sibling_state = to_cuda(sibling_state, device)

            flag_sibling = False
            for q_index in range(len(queue_decode)):
                if (head <= len(queue_decode)) and (q_index < head - 1) and (queue_decode[q_index]["parent"] == queue_decode[head - 1]["parent"]) and (queue_decode[q_index]["child_index"] < queue_decode[head - 1]["child_index"]):
                    flag_sibling = True
                    sibling_index = q_index
            if flag_sibling:
                sibling_state = queue_decode[sibling_index]["s"][1]

            if head == 1:
                prev_word = torch.tensor([form_manager.get_symbol_idx(
                    form_manager.start_token)], dtype=torch.long)
            else:
                prev_word = torch.tensor(
                    [form_manager.get_symbol_idx('(')], dtype=torch.long)

            prev_word = to_cuda(prev_word, device)
            i_child = 1
            if not use_beam_search:
                while True:
                    prediction, (curr_c, curr_h), _ = model.decoder.decode_step(tgt_batch_size=1,
                                                                 dec_single_input=prev_word,
                                                                 dec_single_state=s,
                                                                 memory=enc_outputs,
                                                                 parent_state=parent_h,
                                                                 oov_dict=oov_dict,
                                                                 enc_batch=enc_w_list)
                    s = (curr_c, curr_h)
                    prev_word = torch.log(prediction + 1e-31)
                    prev_word = prev_word.argmax(1)

                    if int(prev_word[0]) == form_manager.get_symbol_idx(form_manager.end_token) or t.num_children >= max_dec_seq_length:
                        break
                    elif int(prev_word[0]) == form_manager.get_symbol_idx(form_manager.non_terminal_token):
                        queue_decode.append({"s": (s[0].clone(), s[1].clone()), "parent": head, "child_index": i_child, "t": Tree()})
                        t.add_child(int(prev_word[0]))
                    else:
                        t.add_child(int(prev_word[0]))
                    i_child = i_child + 1
            else:
                topk = 1
                # decoding goes sentence by sentence
                assert(graph_node_embedding.size(0) == 1)
                beam_search_generator = DecoderStrategy(
                    beam_size=beam_size, vocab=form_manager, decoder=model.decoder, rnn_type="lstm", use_copy=True, use_coverage=False)
                decoded_results = beam_search_generator.beam_search_for_tree_decoding(decoder_initial_state=(s[0], s[1]),
                                                                                          decoder_initial_input=prev_word,
                                                                                          parent_state=parent_h,
                                                                                          graph_node_embedding=enc_outputs,
                                                                                          rnn_node_embedding=rnn_node_embedding,
                                                                                          device=device,
                                                                                          topk=topk,
                                                                                          oov_dict=oov_dict,
                                                                                          enc_batch=enc_w_list)
                generated_sentence = decoded_results[0][0]
                for node_i in generated_sentence:
                    if int(node_i.wordid.item()) == form_manager.get_symbol_idx(form_manager.non_terminal_token):
                        queue_decode.append({"s": (node_i.h[0].clone(), node_i.h[1].clone(
                        )), "parent": head, "child_index": i_child, "t": Tree()})
                        t.add_child(int(node_i.wordid.item()))
                        i_child = i_child + 1
                    elif int(node_i.wordid.item()) != form_manager.get_symbol_idx(form_manager.end_token) and \
                            int(node_i.wordid.item()) != form_manager.get_symbol_idx(form_manager.start_token) and \
                            int(node_i.wordid.item()) != form_manager.get_symbol_idx('('):
                        t.add_child(int(node_i.wordid.item()))
                        i_child = i_child + 1

            head = head + 1
        for i in range(len(queue_decode)-1, 0, -1):
            cur = queue_decode[i]
            queue_decode[cur["parent"] -
                         1]["t"].children[cur["child_index"]-1] = cur["t"]
        return queue_decode[0]["t"].to_list(form_manager)

    def _build_rnn(self, rnn_type, input_size, emb_size, hidden_size, dropout_input, use_sibling, device):
        """_build_rnn : how the rnn unit should be build.
        """
        rnn = TreeDecodingUnit(input_size, emb_size, hidden_size, dropout_input, use_sibling, self.embeddings)

        return rnn

    def forward(self, g, tgt_tree_batch=None, oov_dict=None):
        params = self._extract_params(g)
        params['tgt_tree_batch'] = tgt_tree_batch
        params["oov_dict"] = oov_dict
        return self._run_forward_pass(**params)

    def _extract_params(self, graph_list):
        """

        Parameters
        ----------
        g: GraphData

        Returns
        -------
        params: dict
        """
        batch_data_dict = graph_list.batch_node_features
        graph_node_emb = batch_data_dict["node_emb"]
        
        # [s_g.node_features["node_emb"] for s_g in graph_list]
        rnn_node_emb = batch_data_dict["rnn_emb"]

        graph_node_mask = (batch_data_dict["token_id"] != 0).squeeze(-1).float() - 1

        if self.use_copy:
            src_seq_ret = graph_list.batch_node_features["token_id_oov"]
        else:
            src_seq_ret = None

        return {
            "graph_node_embedding": graph_node_emb,
            "graph_node_mask": graph_node_mask,
            "rnn_node_embedding": rnn_node_emb,
            "graph_level_embedding": None,
            "graph_edge_embedding": None,
            "graph_edge_mask": None,
            "enc_batch": src_seq_ret.long() if self.use_copy else None
        }

def create_mask(x, N, device=None):
    x = x.data
    mask = np.zeros((x.size(0), N))
    for i in range(x.size(0)):
        mask[i, :x[i]] = 1
    return torch.Tensor(mask).to(device)


class TreeDecodingUnit(nn.Module):
    def __init__(self, input_size, emb_size, hidden_size, dropout_input, use_sibling, dec_embeddings):
        super(TreeDecodingUnit, self).__init__()
        self.hidden_size = hidden_size
        self.emb_size = emb_size
        self.embedding = dec_embeddings
        self.dropout = nn.Dropout(dropout_input)

        self.lstm = nn.LSTMCell(
            emb_size + hidden_size * (2 if use_sibling else 1), hidden_size)
        self.use_sibling = use_sibling

    def forward(self, input_src, prev_c, prev_h, parent_h, sibling_state):

        src_emb = self.embedding(input_src)
        src_emb = self.dropout(src_emb)
        if self.use_sibling:
            input_single_step = torch.cat(
                (src_emb, parent_h, sibling_state), 1)
        else:
            input_single_step = torch.cat((src_emb, parent_h), 1)
        prev_cy, prev_hy = self.lstm(input_single_step, (prev_c, prev_h))
        return prev_cy, prev_hy, input_single_step


def get_dec_batch(dec_tree_batch, batch_size, device, form_manager):
    queue_tree = {}
    for i in range(1, batch_size+1):
        queue_tree[i] = []
        queue_tree[i].append(
            {"tree": dec_tree_batch[i-1], "parent": 0, "child_index": 1})

    cur_index, max_index = 1, 1
    dec_batch = {}
    # max_index: the max number of sequence decoder in one batch
    while (cur_index <= max_index):
        max_w_len = -1
        batch_w_list = []
        for i in range(1, batch_size+1):
            w_list = []
            if (cur_index <= len(queue_tree[i])):
                t = queue_tree[i][cur_index - 1]["tree"]

                for ic in range(t.num_children):
                    if isinstance(t.children[ic], Tree):
                        w_list.append(4)
                        queue_tree[i].append(
                            {"tree": t.children[ic], "parent": cur_index, "child_index": ic + 1})
                    else:
                        w_list.append(t.children[ic])
                if len(queue_tree[i]) > max_index:
                    max_index = len(queue_tree[i])
            if len(w_list) > max_w_len:
                max_w_len = len(w_list)
            batch_w_list.append(w_list)
        dec_batch[cur_index] = torch.zeros(
            (batch_size, max_w_len + 2), dtype=torch.long)
        for i in range(batch_size):
            w_list = batch_w_list[i]
            if len(w_list) > 0:
                for j in range(len(w_list)):
                    dec_batch[cur_index][i][j+1] = w_list[j]
                # add <S>, <E>
                if cur_index == 1:
                    dec_batch[cur_index][i][0] = 1
                else:
                    dec_batch[cur_index][i][0] = form_manager.get_symbol_idx(
                        '(')
                dec_batch[cur_index][i][len(w_list) + 1] = 2

        dec_batch[cur_index] = to_cuda(dec_batch[cur_index], device)
        cur_index += 1

    return dec_batch, queue_tree, max_index


import torch
import torch.nn as nn
import dgl

from ..base import PoolingBase
from .....data.data import from_batch


class MaxPooling(PoolingBase):
    r"""Apply max pooling over the nodes in the graph.

    .. math::
        r^{(i)} = \max_{k=1}^{N_i}\left( x^{(i)}_k \right)
    """
    def __init__(self, dim=None, use_linear_proj=False):
        super(MaxPooling, self).__init__()
        if use_linear_proj:
            assert dim is not None, "dim should be specified when use_linear_proj is set to True"
            self.linear = nn.Linear(dim, dim, bias=False)
        else:
            self.linear = None

    def forward(self, graph, feat):
        r"""Compute max pooling.

        Parameters
        ----------
        graph : GraphData
            The graph data.
        feat : str
            The feature field name.

        Returns
        -------
        torch.Tensor
            The output feature.
        """
        graph_list = from_batch(graph)
        output_feat = []
        for g in graph_list:
            feat_tensor = g.node_features[feat]
            if self.linear is not None:
                feat_tensor = self.linear(feat_tensor)

            output_feat.append(torch.max(feat_tensor, dim=0)[0])

        output_feat = torch.stack(output_feat, 0)

        return output_feat



import copy
import warnings

import numpy as np
import torch

from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase


class Accuracy(EvaluationMetricBase):
    """
        Calculate precision, recall, F1 for each labels

    Parameters
    ----------
    metrics: list
        Indicate the metric for the class to return.
        Note that each metric must be one of ``precision``, ``recall``, ``F1``, ``accuracy``.
        And the results' order is the same as the metrics.
    """

    def __init__(self, metrics):
        super().__init__()
        if isinstance(metrics, list):
            for metric in metrics:
                if metric not in ["precision", "recall", "F1", "accuracy"]:
                    raise TypeError(
                        "argument metric must be list of str containing 'precision', 'recall', 'F1', 'accuracy'")
        self.metrics = metrics

    def calculate_scores(self, ground_truth, predict, average=None, zero_division="warning", sample_weight=None):
        """
            The function to calculate the expected metrics for each labels

        Parameters
        ----------
        ground_truth: torch.Tensor
            Ground truth (correct) target values, 1d tensor
        predict: torch.Tensor
            The predicted target values generated by classifier, 1d tensor
        average: string or None, [None (default), 'micro', 'macro', 'weighted']
            If set ``None``, it will return the scores for each class. Otherwise, it will be reduced \
                by the strategy as follows:

            ``'micro'``:
                Calculate metrics globally by counting the total true positives,
                false negatives and false positives.

            ``'macro'``:
                Calculate metrics for each label, and calculate the unweighted
                average values.  This does not take label imbalance into account.

            ``'weighted'``:
                Calculate metrics for each label, and calculate the weighted
                average value. Note that the weight is the number of the true
                instances for each label.

        zero_division: "warning", 0, 1, default="warning"
                Sets the value to return when there is a zero division.

                If set to "warning", this acts as 0, but warnings are also raised.

        sample_weight: None
            The sample weight. It is not implemented yet.

        Returns
        -------
        scores: list[object]
            Return the expected metrics initialized in init function in ``precision``, ``recall``, ``F1`` order
        """
        ground_truth_np, predict_np = self._check_available(ground_truth, predict, zero_division)

        # calculate accuracy
        scores = ground_truth_np == predict_np
        accuracy_score = np.average(scores)

        if self.metrics is ["precision"]:
            return [accuracy_score]

        mcm = self._calculate_confusion_matrix(ground_truth=ground_truth_np, predict=predict_np)

        tp_sum = mcm[:, 1, 1]
        pred_sum = tp_sum + mcm[:, 0, 1]
        gt_sum = tp_sum + mcm[:, 1, 0]

        if average == 'micro':
            tp_sum = np.array([tp_sum.sum()])
            pred_sum = np.array([pred_sum.sum()])
            gt_sum = np.array([gt_sum.sum()])

        # calculate precision and recall
        precision = self._prf_divide(tp_sum, pred_sum, zero_division=zero_division)
        recall = self._prf_divide(tp_sum, gt_sum, zero_division=zero_division)

        # calculate F_beta
        beta2 = 1 ** 2  # note: only F1 here
        denominator = beta2 * precision + recall

        denominator[denominator == 0.] = 1  # avoid division by 0
        f_score = (1 + beta2) * precision * recall / denominator

        if average == "weighted":
            weighted = gt_sum
        else:
            weighted = None

        if average is not None:
            precision = np.average(precision, weights=weighted)
            recall = np.average(recall, weights=weighted)
            f_score = np.average(f_score, weights=weighted)
        scores = []

        for metric_name in self.metrics:
            if metric_name is "precision":
                scores.append(precision)
            elif metric_name is "recall":
                scores.append(recall)
            elif metric_name is "F1":
                scores.append(f_score)
            elif metric_name is "accuracy":
                scores.append(accuracy_score)
            else:
                raise NotImplementedError()
        return scores

    @staticmethod
    def _prf_divide(numerator, denominator, zero_division):
        """
            The function performs division and handles zero-division situations.

        Parameters
        ----------
        numerator: numpy.ndarray
        denominator: numpy.ndarray
        zero_division: "warning", 0, 1, default="warning"
            Sets the value to return when there is a zero division.

            If set to "warning", this acts as 0, but warnings are also raised.
        Returns
        -------
        results: numpy.ndarray
            The division results.

        """
        zero_mask = denominator == 0.
        denominator_cp = copy.deepcopy(denominator)
        denominator_cp[zero_mask] = 1.
        ret = numerator / denominator_cp
        if np.sum(zero_mask) == 0:
            return ret
        ret[zero_mask] = 0. if zero_division in ["warning", 0] else 1.
        if zero_division is "warning":
            warnings.warn("zero division encountered")
        return ret

    @staticmethod
    def _check_available(ground_truth, predict, zero_division):
        """
            The function to check the parameters.
            If all tests are passed, it will convert the tensor to numpy.

        Parameters
        ----------
        ground_truth: Any
        predict: Any
        zero_division: Any

        Returns
        -------
        ground_truth: numpy.ndarray
            numpy version of tensor ground_truth
        predict: numpy.ndarray
            numpy version of tensor predict

        Raises
        -------
        TypeError: TypeError
        ValueError: ValueError
        """
        if not isinstance(ground_truth, torch.Tensor):
            raise TypeError("argument ground_truth must be torch.tensor")
        if not isinstance(predict, torch.Tensor):
            raise TypeError("argument predict must be torch.tensor")
        if ground_truth.dtype not in [torch.int, torch.int8, torch.int16, torch.int32, torch.int64]:
            raise TypeError("argument ground_truth must be int tensor")
        if predict.dtype not in [torch.int, torch.int8, torch.int16, torch.int32, torch.int64]:
            raise TypeError("argument predict must be int tensor")
        if len(ground_truth.shape) != 1:
            raise TypeError("argument ground_truth must be 1d tensor")
        if len(predict.shape) != 1:
            raise TypeError("argument predict must be 1d tensor")
        if ground_truth.shape[0] != predict.shape[0]:
            raise ValueError("argument ground_truth and predict must be the same shape")

        zero_division_ok = False
        if isinstance(zero_division, str) and zero_division is "warning":
            zero_division_ok = True
        elif isinstance(zero_division, (int, float)) and zero_division in [0, 1]:
            zero_division_ok = True

        if not zero_division_ok:
            raise ValueError("argument zero_division must be in ['warning', 0, 1]")

        return ground_truth.numpy(), predict.numpy()

    def _calculate_confusion_matrix(self, ground_truth, predict):
        """
            The function to calculate the confusion matrix for multi-class inputs.
            The labels will be collected and relabeled. (eg: [1, 2, 3] --> [0, 1, 2])

            In multi-class confusion matrix :math:`MCM`, the count of true negatives is
            :math:`MCM_{:,0,0}`, false positives is :math:`MCM_{:,0,1}`, false negatives
            is :math:`MCM_{:,1,0}` and true positive is :math:`MCM_{:,1,1}`.

        Parameters
        ----------
        ground_truth: numpy.ndarray
        predict: numpy.ndarray

        Returns
        -------
        confusion_matrix: numpy.ndarray
            The confusion matrix which has the shape: [num_labels, 2, 2]
        """
        # select all labels, remove duplicates and sort
        unique_labels = sorted(self._get_unique_labels(ground_truth, predict))

        # do relabeling
        ground_truth_transformed = np.searchsorted(unique_labels, ground_truth)
        predict_transformed = np.searchsorted(unique_labels, predict)

        # the number of labels after relabeling
        n_labels = len(unique_labels)

        tp = ground_truth_transformed == predict_transformed
        tp_bins = ground_truth_transformed[tp]
        tp_sum = np.bincount(tp_bins, weights=None, minlength=n_labels)
        pred_sum = np.bincount(predict_transformed, minlength=n_labels)
        gt_sum = np.bincount(ground_truth_transformed, minlength=n_labels)
        fp = pred_sum - tp_sum
        fn = gt_sum - tp_sum
        tp = tp_sum
        tn = ground_truth_transformed.shape[0] - tp - fp - fn
        return np.array([tn, fp, fn, tp]).T.reshape(-1, 2, 2)

    @staticmethod
    def _get_unique_labels(*lists):
        """
            find the unique elements in the given lists

        Parameters
        ----------
        lists: [numpy.ndarray]
            List of lists which contain labels.
        Returns
        -------
        unique_labels: numpy.ndarray
            It has unique labels encountered in the ``lists``.
        """
        ret = []
        for li in lists:
            unique_li = np.unique(li)
            ret.extend(unique_li.tolist())
        ret = list(set(ret))
        return np.array(ret)
